import matplotlib.pyplot as plt
import scipy as sp
import numpy as np
import torch


def estimate_upper_control_limit(data, alpha=0.05, band_width="scott"):
    """
    Estimate upper control limit (UCL) using kernel density estimation (KDE) with a gaussian kernel
    :param data: input numpy array of statistics
    :param alpha: significance level for finding the (1-alpha) percentile
    :param band_width: if "standard" uses elbow rule, otherwise can be a float value to be used as bandwidth
    :return: value of the empirical cumulative distribution function corresponding to (1-alpha)-upper percentile
    """
    bw = sp.stats.gaussian_kde(data, bw_method=band_width).covariance_factor() * data.std()
    return sp.optimize.brentq(lambda x: sum(sp.stats.norm.cdf((x - data) / bw)) / len(data) - (1 - alpha), 0, 100000)


def get_grads(x, trained_model):
    inp = x.unsqueeze(0).requires_grad_(True)
    output = trained_model.encoder(inp)  # trained model should have the encoding layers saved as "encoder"
    gradient = torch.autograd.grad(output.mean(), inp)[0][0].data.numpy()
    return gradient


def integrated_gradients(model, inp, baseline, approximation_method="monte_carlo", steps=20, samples=1000):
    # If baseline is not provided, start with an array of zeros
    if baseline is None:
        baseline = np.zeros(len(inp)).astype(np.float32)
    else:
        baseline = baseline

    baseline = np.array(baseline).reshape(-1, len(inp))
    inp = np.array(inp).reshape(-1, len(inp))

    if approximation_method == "trapezoidal":
        path_inputs = [baseline + (i / steps) * (inp - baseline) for i in range(steps + 1)]
        grads = []
        for elem in path_inputs:
            grad = get_grads(torch.Tensor(elem[0]), model)
            grads.append(np.array(grad))
        delta = 1 / steps
        integrated_grads = delta * np.sum((inp - baseline) * grads, axis=0)

    elif approximation_method == "monte_carlo":
        random_samples = np.random.uniform(0, 1, (samples, len(inp)))
        random_vectors = baseline + (inp - baseline) * random_samples
        grads = []
        for i in range(samples):
            grad = get_grads(torch.Tensor(random_vectors[i, :]), model)
            grads.append(np.array(grad))
        integrated_grads = (inp - baseline) * np.mean(grads, axis=0)

    else:
        raise ValueError(
            "Invalid approximation method: {}, use either monte_carlo or trapezoidal".format(approximation_method))

    return integrated_grads


def plot_multivariate_control_charts(t2_scores, t2_ucl, spe_scores, spe_ucl, log_scale=False):
    if log_scale:
        t2_scores = np.log(t2_scores)
        t2_ucl = np.log(t2_ucl)
        spe_scores = np.log(spe_scores)
        spe_ucl = np.log(spe_ucl)

    fig, axs = plt.subplots(2, 1, figsize=(8, 8))
    fig.subplots_adjust(hspace=0.5, bottom=0.1)
    # T^2 chart
    axs[0].plot(t2_scores, marker='', color='black', linewidth=0.8, alpha=1)
    axs[0].axhline(y=t2_ucl, color='red', label="UCL", alpha=0.6)
    axs[0].grid(False)
    if log_scale:
        axs[0].set_ylabel("$T^2$ statistic (log)")
    else:
        axs[0].set_ylabel("$T^2$ statistic")
    axs[0].set_xlabel("Observations")
    axs[0].legend(loc="upper left")
    axs[0].set_title("$T^2$ Control Chart", loc='center', fontweight=1)

    # SPE chart
    axs[1].plot(spe_scores, marker='', color='black', linewidth=0.8, alpha=1)
    axs[1].axhline(y=spe_ucl, color='red', label="UCL", alpha=0.6)
    axs[1].grid(False)
    if log_scale:
        axs[1].set_ylabel("SPE statistic (log)")
    else:
        axs[1].set_ylabel("SPE statistic")
    axs[1].set_xlabel("Observations")
    axs[1].legend(loc="upper left")
    axs[1].set_title("SPE Control Chart", loc='center', fontweight=1)
    plt.show()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
import torch
from autoencoders.vae.vae import VAE
from autoencoders.vae.training import TrainingModel
from autoencoders.oae.utils import estimate_upper_control_limit, plot_multivariate_control_charts


class ProcessMonitoring:

    def __init__(self, alpha=0.05):
        self.alpha = alpha
        self.scaler = StandardScaler()
        self.columns = None
        self.model = None
        self.trained_model = None
        self.losses = None
        self.x_encoded_phase1 = None
        self.inverted_cov_matrix = None
        self.mean_vector = None
        self.t2_scores_phase1 = None
        self.ucl_t2 = None
        self.residuals_phase1 = None
        self.ucl_spe = None
        self.t2_scores_phase2 = None
        self.residuals_phase2 = None

    def fit(self, data, encoding_layers, orthogonality_regularization=0.1, norm="fro", nr_epochs=100, patience=10, batch_size=100, verbose=True):
        """
        training autoencoders and finding upper control limits
        :param data: phase 1 data, pandas DataFrame
        :param encoding_layers: layers of the encoder newtork (decoder is symmetrical), e.g., [input_dim, hidden_dim, ..., encoding_dim]
        :param orthogonality_regularization: float value
        :param norm: norm type, can be "fro" (Frobenius), "l1" or "l2"
        :param nr_epochs: max number of epochs
        :param patience: for early stopping, number of epochs without improvement tolerated
        :param batch_size: int
        :param verbose: whether to print the training history or not
        :return:
        """
        # standardizing phase one data
        data = pd.DataFrame(self.scaler.fit_transform(data))

        # initializing model
        self.model = VAE(encoding_layers=encoding_layers, penalty=orthogonality_regularization, norm_type=norm, )
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.model = self.model.to(device)
        x = torch.Tensor(data.values)

        # training model
        trainer = TrainingModel(learning_rate=1e-3, batch_size=batch_size, val_size=0.20, verbose=verbose)
        trainer.create_datasets(train_data=x)
        self.trained_model, self.losses = trainer.train_model(self.model, patience=patience, n_epochs=nr_epochs)

        # obtaining encoded and reconstructed features
        self.x_encoded_phase1, x_reconstructed = self.trained_model(x)
        self.x_encoded_phase1 = self.x_encoded_phase1.cpu().detach().numpy()
        x_reconstructed = pd.DataFrame(x_reconstructed.cpu().detach().numpy())

        # Hotelling T^2 control chart
        self.inverted_cov_matrix = np.linalg.inv(np.cov(self.x_encoded_phase1.T))
        self.mean_vector = np.mean(self.x_encoded_phase1, axis=0)
        self.t2_scores_phase1 = np.array([(sample - self.mean_vector).T @ self.inverted_cov_matrix @ (sample - self.mean_vector) for sample in self.x_encoded_phase1])
        self.ucl_t2 = estimate_upper_control_limit(self.t2_scores_phase1)

        # SPE control chart
        data.reset_index(drop=True, inplace=True)
        self.columns = np.arange(0, data.shape[1])
        self.residuals_phase1 = ((data.subtract(x_reconstructed, axis=1)) ** 2).sum(axis=1)
        self.ucl_spe = estimate_upper_control_limit(self.residuals_phase1)

        print(f"Training complete! Upper control limits are: {self.ucl_t2:4.2f} (T^2) and {self.ucl_spe:.2f} (SPE)")

    def plot_phase1(self, log_scale=False):
        """
        finds UCLs and plots control charts
        :param log_scale: whether to plot the values on a log scale
        :return: classification into in-control and out-of-control with T^2 and SPE
        """
        plot_multivariate_control_charts(t2_scores=self.t2_scores_phase1, t2_ucl=self.ucl_t2,
                                         spe_scores=self.residuals_phase1, spe_ucl=self.ucl_spe, log_scale=log_scale)
        t2_out_of_control = self.t2_scores_phase1 > self.ucl_t2
        spe_out_of_control = self.residuals_phase1 > self.ucl_spe

        return {"T2": t2_out_of_control.astype(int), "SPE": spe_out_of_control.astype(int)}

    def plot_phase2(self, data, log_scale=False):
        """
        finds UCLs and plots control charts
        :param data: phase 2 data
        :param log_scale: whether to plot the values on a log scale
        :return: classification into in-control and out-of-control with T^2 and SPE
        """
        data = pd.DataFrame(self.scaler.transform(data))
        x = torch.Tensor(data.values)

        x_encoded, x_reconstructed = self.trained_model(x)
        x_encoded = x_encoded.cpu().detach().numpy()
        x_reconstructed = pd.DataFrame(x_reconstructed.cpu().detach().numpy())
        self.t2_scores_phase2 = np.array([(sample - self.mean_vector).T @ self.inverted_cov_matrix @ (sample - self.mean_vector) for sample in x_encoded])
        self.residuals_phase2 = ((data.subtract(x_reconstructed, axis=1)) ** 2).sum(axis=1)

        plot_multivariate_control_charts(t2_scores=self.t2_scores_phase2, t2_ucl=self.ucl_t2,
                                         spe_scores=self.residuals_phase2, spe_ucl=self.ucl_spe, log_scale=log_scale)
        t2_out_of_control = self.t2_scores_phase2 > self.ucl_t2
        spe_out_of_control = self.residuals_phase2 > self.ucl_spe

        return {"T2": t2_out_of_control.astype(int), "SPE": spe_out_of_control.astype(int)}

    def plot_correlation_matrix(self):
        """
        Plot the covariance matrix of the encoded features from phase 1
        """
        # Compute the covariance matrix of the encoded features
        cov = np.corrcoef(self.x_encoded_phase1.T)

        # Determine step size for ticks
        n_ticks = len(cov)
        step = max(1, n_ticks // 10)

        # Plot the covariance matrix
        fig, ax = plt.subplots()
        im = ax.imshow(cov, cmap="viridis")
        ax.set_xticks(np.arange(0, n_ticks, step))
        ax.set_yticks(np.arange(0, n_ticks, step))
        ax.set_xticklabels(np.arange(1, n_ticks + 1, step))
        ax.set_yticklabels(np.arange(1, n_ticks + 1, step))
        ax.set_title("Correlation Matrix of Encoded Features")
        fig.colorbar(im)
        plt.show()

    def spe_contribution_plots(self, ooc_observation):
        data = pd.DataFrame(self.scaler.transform(np.array(ooc_observation).reshape(1, -1)))
        x = torch.Tensor(data.values)
        _, x_reconstructed = self.trained_model(x.reshape(1, -1))
        x_reconstructed = x_reconstructed.cpu().detach().numpy()
        residuals = np.array(x - x_reconstructed) ** 2
        plt.bar(self.columns, residuals[0], color="r", alpha=0.6)
        plt.title("SPE Contribution Plots")
        plt.xlabel("Variables")
        plt.ylabel("Squared residuals")
        plt.show()



"""
Unit tests go here
"""
import numpy as np
import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from base._src.helpers import (
    flatten_list,
    get_df_with_sample_id,
    get_num_of_PCs_to_retain,
    multiply_matrices,
    standardize_and_PCA,
)


def test_flatten_list():
    assert flatten_list([[1, 1], [2, 2], [3, 3]]) == [1, 1, 2, 2, 3, 3]
    assert flatten_list([[], []]) == []
    assert flatten_list([[]]) == []


def test_get_df_with_sample_id():
    df_test_input = pd.DataFrame({"x1": [1, 2, 2, 3, 3, 4, 5]})
    df_expected1 = pd.DataFrame(
        {"sample_id": [1, 2, 3, 4, 5, 6, 7], "x1": [1, 2, 2, 3, 3, 4, 5]}
    )
    df_expected2 = pd.DataFrame(
        {"sample_id": [1, 1, 2, 2, 3, 3, 4], "x1": [1, 2, 2, 3, 3, 4, 5]}
    )
    df_expected3 = pd.DataFrame(
        {"sample_id": [1, 1, 1, 1, 1, 1, 1], "x1": [1, 2, 2, 3, 3, 4, 5]}
    )
    df_output1 = get_df_with_sample_id(df_test_input, n_sample_size=1)
    df_output2 = get_df_with_sample_id(df_test_input, n_sample_size=2)
    df_output3 = get_df_with_sample_id(df_test_input, n_sample_size=7)

    assert_frame_equal(df_output1, df_expected1, check_dtype=False)
    assert_frame_equal(df_output2, df_expected2, check_dtype=False)
    assert_frame_equal(df_output3, df_expected3, check_dtype=False)
    with pytest.raises(Exception):
        get_df_with_sample_id(df_test_input, n_sample_size=0)


def test_multiply_matrices():
    sigma = np.array(
        [
            [1, 0.7, 0.9, 0.3, 0.2, 0.3],
            [0.7, 1, 0.8, 0.1, 0.4, 0.2],
            [0.9, 0.8, 1, 0.1, 0.2, 0.1],
            [0.3, 0.1, 0.1, 1, 0.2, 0.1],
            [0.2, 0.4, 0.2, 0.2, 1, 0.1],
            [0.3, 0.2, 0.1, 0.1, 0.1, 1],
        ]
    )
    delta_output = np.sqrt(
        multiply_matrices(np.array([1] * 6), np.linalg.inv(sigma), np.array([1] * 6))
    )
    delta_expected = 1.86
    assert np.round(delta_output, 2) == delta_expected


@pytest.mark.parametrize(
    "variance_explain_min, expected_num_of_PCs_to_retain",
    [[0.1, 1], [0.55, 2], [0.9, 3], [0.95, 3]],
)
def test_get_num_of_PCs_to_retain(
    dataframe_for_PCAModel_phase1, variance_explain_min, expected_num_of_PCs_to_retain
):
    _, PCA_object, _ = standardize_and_PCA(df=dataframe_for_PCAModel_phase1)
    num_of_PCs_to_retain, _ = get_num_of_PCs_to_retain(
        PCA_object=PCA_object, PC_variance_explained_min=variance_explain_min
    )
    assert num_of_PCs_to_retain == expected_num_of_PCs_to_retain



from torch import nn
import torch
import torch.nn.functional as F


class AutoEncoder(nn.Module):

    def __init__(self, encoding_layers, penalty=0.1, norm_type="l2"):
        """
        :param encoding_layers: input dimension is the first item of the list, encoding dimension is the last item of the layers list (decoder is symmetric)
        :param penalty
        :param norm_type: either "frobenius" or "l2"
        """

        super().__init__()
        self.penalty = penalty
        self.norm_type = norm_type

        # Encoding layers
        encoding = []
        for i in range(len(encoding_layers) - 1):
            encoding.append(nn.Linear(encoding_layers[i], encoding_layers[i + 1]))
            encoding.append(nn.ReLU(inplace=True))
            encoding.append(nn.BatchNorm1d(encoding_layers[i + 1]))

        # Decoding layers
        decoding_layers = encoding_layers[::-1]
        decoding = []
        for i in range(len(decoding_layers) - 1):
            decoding.append(nn.Linear(decoding_layers[i], decoding_layers[i + 1]))
            decoding.append(nn.ReLU(inplace=True))
            decoding.append(nn.BatchNorm1d(decoding_layers[i + 1]))

        self.encoder = nn.Sequential(*encoding)
        self.decoder = nn.Sequential(*decoding)

    def forward(self, x):
        embeddings = []
        for layer in self.encoder:
            x = layer(x)
            embeddings.append(x)

        x_encoded = embeddings[-1]
        x_reconstructed = self.decoder(x_encoded)
        return x_encoded, x_reconstructed

    def loss_function(self, compression, reconstruction, original):
        """
        Ortho-loss to learn uncorrelated features. The penalty term computed as the Frobenius or L2 norm of the difference
        between the dot product of the encoded features and the identity matrix may be negative if the dot product is
        greater than the identity matrix. By squaring the penalty term, we ensure that it is always positive and has a
        greater impact on the overall loss function, which encourages the encoded features to be more orthogonal.
        :param compression: encoded features from the bottleneck
        :param reconstruction: reconstructed features from the last layer
        :param original: input features
        :return: loss value
        """
        mse_loss = F.mse_loss(reconstruction, original)

        # Compute penalty term
        dot_product = torch.matmul(compression.T, compression)
        if self.norm_type == 'frobenius':
            penalty_loss = torch.norm(dot_product - torch.eye(dot_product.shape[0]), p='fro') ** 2
        elif self.norm_type == 'l2':
            penalty_loss = torch.norm(dot_product - torch.eye(dot_product.shape[0]), p=2) ** 2
        else:
            raise ValueError("Invalid norm type: {}, use either frobenius or l2 norms".format(self.norm_type))
        loss = mse_loss + self.penalty * penalty_loss
        return loss



univariate_phase1_color = "green"
univariate_phase2_color = "black"
multivariate_phase1_color = "green"
multivariate_phase2_color = "black"
outlier_marker_color = "red"
UCL_color = "red"
LCL_color = "red"
single_phase_line_color = "black"


def get_outside_CL_marker_size(N_samples: int):
    return 15 if N_samples > 500 else 25


def get_line_width(N_samples: int):
    return 1 if N_samples > 500 else 1.5



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_Analysis.ipynb.

# %% auto 0
__all__ = ['ACM_analysis', 'PACM_analysis', 'Analysis']

# %% ../nbs/03_Analysis.ipynb 2
import pandas as pd
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from lag import *
from corrmat import *
from TEP import *

# %% ../nbs/03_Analysis.ipynb 4
def ACM_analysis(X: pd.DataFrame, # Raw data to perform analysis on
                lag: int # Number of lags to investigate
                )-> dict: # Dict with eigenvalues
    
    res_dict = {}
    
    for i in range(lag+1):
        eigenValues, eigenVectors = np.linalg.eig(ACM(X,i))
        idx_abs = abs(eigenValues).argsort()[::-1]
        eigenValues = abs(eigenValues)[idx_abs]
        eigenVectors = np.real(eigenVectors)[idx_abs]
        res_dict['EigenValues',i] = eigenValues
        res_dict['EigenVectors',i] = eigenVectors
    return res_dict


# %% ../nbs/03_Analysis.ipynb 5
def PACM_analysis(X: pd.DataFrame, # Raw data to perform analysis on
                lag: int # Number of lags to investigate
                )-> dict: # Dict with eigenvalues
    res_dict = {}
    for i in range(lag+1):
        eigenValues, eigenVectors = np.linalg.eig(PACM(X,i))
        idx_abs = abs(eigenValues).argsort()[::-1]
        eigenValues = abs(eigenValues)[idx_abs]
        eigenVectors = np.real(eigenVectors)[idx_abs]
        res_dict['EigenValues', i] = eigenValues
        res_dict['EigenVectors', i] = eigenVectors
    return res_dict

# %% ../nbs/03_Analysis.ipynb 6
class Analysis:
    def __init__(self,lag):
        self.lag = lag
        self.imposter = None
        self.PACM = None
        self.PACM_imposter = None
        self.ACM = None
        self.ACM_imposter = None
    
    def show_plots(self):
        
        fig, axs = plt.subplots(1,2, figsize = (10,5))
        
        for i in range(self.lag+1):
            axs[0].scatter(i, self.ACM['EigenValues',i][0], c = 'blue')
            axs[0].scatter(i, self.ACM_imposter['EigenValues',i][0], c = 'red', marker = 'x')
            axs[0].set_title('Autocorrelation Matrix Eigenvalues', fontweight = 'bold')
            axs[0].set_xlabel('Lag')
            axs[0].set_ylabel('Largest Absolute Eigenvalue')
            axs[0].legend(['Data','Imposter'])
            
            axs[1].scatter(i, self.PACM['EigenValues',i][0], c = 'blue')
            axs[1].scatter(i, self.PACM_imposter['EigenValues',i][0], c = 'red', marker = 'x')
            axs[1].set_title('Partial Autocorrelation Matrix Eigenvalues', fontweight = 'bold')
            axs[1].set_xlabel('Lag')
            axs[1].set_ylabel('Largest Absolute Eigenvalue')
            axs[1].legend(['Data','Imposter'])

        fig.savefig('output.png')
    
    def fit(self, X : pd.DataFrame):
        self.imposter = imposter_matrix(X)
        self.PACM = PACM_analysis(X, self.lag)
        self.PACM_imposter = PACM_analysis(self.imposter,self.lag)
        self.ACM = ACM_analysis(X, self.lag)
        self.ACM_imposter = ACM_analysis(self.imposter, self.lag)
        return self


    

        



from abc import abstractmethod
from typing import Optional, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.stats import beta, chi2, f

from base._src.constants import get_A_constant, get_B_constant, get_D_constant
from base._src.helpers import (
    ControlChartPlotMixin,
    apply_standardize_and_PCA,
    get_df_with_sample_id,
    get_num_of_PCs_to_retain,
    multiply_matrices,
    standardize_and_PCA,
    vprint,
)


class BaseControlChart(ControlChartPlotMixin):
    def __init__(self, n_sample_size: int):
        self.is_fitted = False
        self.UCL = None
        self.LCL = None
        self.center_line = None
        self.stat_name = None
        self.n_sample_size = n_sample_size
        self.hyperparameters = None

    @abstractmethod
    def fit(self, df_phase1: pd.DataFrame, *args, **kwargs):
        pass

    def _collect_results_df(self, df_stats: pd.DataFrame):
        df_stats_copy = df_stats.copy()
        if self.n_sample_size == 1:
            df_stats_copy = df_stats_copy[[self.input_name, self.stat_name]]
        else:
            df_stats_copy = df_stats_copy[[self.stat_name]]
        df_stats_copy["LCL"] = self.LCL
        df_stats_copy["UCL"] = self.UCL
        df_stats_copy["outside_CL"] = ~df_stats_copy[self.stat_name].between(
            self.LCL, self.UCL, inclusive="neither"
        )
        return df_stats_copy


class XBarChart(BaseControlChart):
    """
    Univariate control chart for monitoring the process mean of a quality characteristic.
    Assumes normally distributed, non-autocorrelated data. This chart goes hand-in-hand with the
    RChart, which estimates process variability.

    Some tips:
    - For phase 1 limits: first establish statistical control in the RChart, as the XBarChart estimates
      are based on the estimate of process variablity. If many samples plot outside control limits, it can
      pay off to investigate the patterns of these rather than each individual point.
      Montgomery 2013, chapter 6.2, p. 239.
    - Sample sizes, n, of at least 4 or 5 are usually enough to ensure robustness to normality assumption.
      Do note, that the RChart is more sensitive to non-normal data than the XBarChart
      Montgomery 2013, chapter 6.2.5, p. 254.
    - Control charts based on an estimate of the standard deviation, sigma, is more suitable for
      sample sizes, n > 10 or 12, or n is variable.
      Montgomery 2013, chapter 6.3, p. 259.
    """

    def __init__(self, n_sample_size: int = 5,
                 variability_estimator: str = "auto",
                 standard_deviations: float = 3
                 ):
        """
        Instantiates the XBarChart object

        :param n_sample_size: number of individual observations in each sample aka rational subgroup size
        :param variability_estimator: the estimator to use for data variablity, either "auto", "range"
        or "std" (standard deviation). Std is preferred if n_sample_size > 10. Defaults to "auto", which
        automatically decides which estimator is most appropriate for the situation.
        """
        super().__init__(n_sample_size=n_sample_size)
        assert self.n_sample_size > 1, "Sample/subgroup size must be greater than 1"
        assert variability_estimator.lower() in ["auto", "range", "std"], (
            "Variance estimator must be one of " '"auto", "range" or "std"' ""
        )
        self.stat_name = "sample_mean"
        self.standard_deviations = standard_deviations
        self.variability_Rbar_or_sbar = None
        self.input_name = None
        self.df_phase1_stats = None
        self.df_phase1_results = None
        self.df_phase2_stats = None
        self.variability_estimator = self._determine_variability_estimator(
            variability_estimator
        )

    def fit(self, df_phase1: pd.DataFrame, *args, **kwargs):
        """
        Estimates the X-bar control limits based on the provided phase 1 data.

        :param df_phase1: a dataframe of "in-control" / healthy / normal data. This is often referred to as phase 1
        :return: the fitted XBarChart object
        """
        df_phase1_copy = df_phase1.copy()
        df_phase1_copy = get_df_with_sample_id(
            df_phase1_copy, n_sample_size=self.n_sample_size
        )
        self.input_name = df_phase1_copy.columns.values[1]
        df_phase1_copy = self._remove_samples_with_only_one_observation(
            df_phase1_copy
        )  # std not defined on singletons
        self.df_phase1_stats = self._get_df_with_sample_means_and_variability(
            df_phase1_copy, grouping_column="sample_id"
        )
        self.center_line = float(
            self.df_phase1_stats[self.stat_name].mean()
        )  # x double bar
        self.variability_Rbar_or_sbar = float(
            self.df_phase1_stats["sample_variability"].mean()
        )  # R bar or s bar

        constant_A_number = 2 if self.variability_estimator == "range" else 3
        variability_constant = get_A_constant(
            A_number=constant_A_number, n_sample_size=self.n_sample_size
        )

        self.LCL = (
            self.center_line - self.standard_deviations * variability_constant * self.variability_Rbar_or_sbar
        )
        self.UCL = (
            self.center_line + self.standard_deviations * variability_constant * self.variability_Rbar_or_sbar
        )
        self.is_fitted = True
        self.df_phase1_results = self.get_phase1_results()
        return self

    def plot_phase1(self):
        """
        Plots phase 1 statistics, in this case the sample averages and the estimated control limits.
        """
        self._plot_single_phase_univariate(self.df_phase1_results)
        plt.title(r"Phase 1 $\bar{X}$-chart")
        plt.ylabel(r"Sample average [$\bar{x}$]")
        plt.xlabel("Sample")

    def plot_phase2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 2 statistics, in this case the sample averages and the estimated control limits.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        df_phase2_copy = df_phase2.copy()
        if self.df_phase2_stats is None:
            df_phase2_copy = get_df_with_sample_id(
                df_phase2_copy, n_sample_size=self.n_sample_size
            )
            self.df_phase2_stats = self._get_df_with_sample_means_and_variability(
                df_phase2_copy, grouping_column="sample_id"
            )
        df_phase2_results = self._collect_results_df(self.df_phase2_stats)
        self._plot_single_phase_univariate(df_phase2_results)
        plt.title(r"Phase 2 $\bar{X}$-chart")
        plt.ylabel(r"Sample average [$\bar{x}$]")
        plt.xlabel("Sample")

    def plot_phase1_and_2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 1 and 2 statistics, in this case the sample averages and the estimated control limits. Phase 1
        and 2 will be displayed in differing colours to easily tell them apart visually.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        df_phase2_copy = df_phase2.copy()
        df_phase2_copy = get_df_with_sample_id(
            df_phase2_copy, n_sample_size=self.n_sample_size
        )

        if self.df_phase2_stats is None:
            # TODO: need to refactor this: the object should not have attributes related to phase 2 data
            self.df_phase2_stats = self._get_df_with_sample_means_and_variability(
                df_phase2_copy, grouping_column="sample_id"
            )

        df_phase2_results = self._collect_results_df(self.df_phase2_stats)
        self._plot_two_phases(self.df_phase1_results, df_phase2_results)
        plt.title(r"Phase 1 and 2 $\bar{X}$-chart")
        plt.ylabel(r"Sample average [$\bar{x}$]")

    def plot_control_chart(
        self, df_phase1: pd.DataFrame = None, df_phase2: pd.DataFrame = None
    ):
        """
        Convenience wrapper that can plot phase 1 or 2 alone as well as both together. The plotted combination is
        inferred from the given dataframes.

        :param df_phase1: a dataframe of "in-control" / healthy / normal data. This is often referred to as phase 1
        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        if df_phase1 is not None and df_phase2 is None:
            self.plot_phase1()
        if df_phase1 is None and df_phase2 is not None:
            self.plot_phase2(df_phase2=df_phase2)
        if df_phase1 is not None and df_phase2 is not None:
            self.plot_phase1_and_2(df_phase2=df_phase2)

    def get_phase1_results(self):
        assert (
            self.is_fitted
        ), "Method parameters not fitted yet. Run .fit() on phase 1 data first."
        self.df_phase1_results = self._collect_results_df(self.df_phase1_stats)
        return self.df_phase1_results

    def get_phase2_results(self, df_phase2: pd.DataFrame):
        df_phase2_results = self._group_samples_and_compute_stats(df_phase2)
        return self._collect_results_df(df_phase2_results)

    def _get_df_with_sample_means_and_variability(
        self, df: pd.DataFrame, grouping_column: str = "sample_id"
    ):
        def variability_fun(grp):
            return (
                max(grp) - min(grp)
                if self.variability_estimator == "range"
                else np.std(grp, ddof=1)
            )

        df_means_and_variabilities = df.groupby(grouping_column).agg(
            sample_mean=pd.NamedAgg(column=self.input_name, aggfunc=np.mean),
            sample_variability=pd.NamedAgg(
                column=self.input_name, aggfunc=variability_fun
            ),
        )
        return df_means_and_variabilities

    def _group_samples_and_compute_stats(self, df: pd.DataFrame):
        df_copy = df.copy()
        df_copy = get_df_with_sample_id(df_copy, n_sample_size=self.n_sample_size)
        df_copy = self._get_df_with_sample_means_and_variability(
            df_copy, grouping_column="sample_id"
        )
        return df_copy

    def _determine_variability_estimator(self, variability_estimator: str):
        if variability_estimator.lower() == "auto":
            return "std" if self.n_sample_size > 10 else "range"
        else:
            return variability_estimator.lower()

    def _remove_samples_with_only_one_observation(
        self, df_with_sample_id: pd.DataFrame
    ):
        """
        This function removes samples with only one observation to guard against 0 or nans
        in the estimation of variability by either range or standard deviation
        :param df_with_sample_id: a dataframe grouped by sample_id
        :return: the same dataframe, but with size 1 samples removed
        """
        df_grouped = df_with_sample_id.groupby("sample_id").count()
        samples_to_keep = df_grouped[df_grouped[self.input_name] > 1].index.values
        return df_with_sample_id[df_with_sample_id["sample_id"].isin(samples_to_keep)]


class RChart(XBarChart):
    """
    Univariate control chart for monitoring the process variation of a quality characteristic.
    Assumes normally distributed, non-autocorrelated data. The R-chart uses an estimate
    of variability based on the in-sample range contrary to the s-chart, which uses the
    standard deviation to estimate variability.
    The R-chart is best suited for samples of size n < 10 or 12.
    """

    def fit(self, df_phase1: pd.DataFrame, *args, **kwargs):
        """
        Estimates the R-chart control limits based on the provided phase 1 data.

        :param df_phase1: a dataframe of "in-control" / healthy / normal data. This is often referred to as phase 1.
        :return: the fitted RChart object
        """
        super().fit(df_phase1=df_phase1)
        self.stat_name = "sample_variability"

        constant_B3_or_D3 = get_B_constant(B_number=3, n_sample_size=self.n_sample_size)
        constant_B4_or_D4 = get_B_constant(B_number=4, n_sample_size=self.n_sample_size)
        if self.variability_estimator == "range":
            constant_B3_or_D3 = get_D_constant(
                D_number=3, n_sample_size=self.n_sample_size
            )
            constant_B4_or_D4 = get_D_constant(
                D_number=4, n_sample_size=self.n_sample_size
            )

        self.LCL = constant_B3_or_D3 * self.variability_Rbar_or_sbar
        self.UCL = constant_B4_or_D4 * self.variability_Rbar_or_sbar
        self.center_line = self.variability_Rbar_or_sbar
        self.df_phase1_results = self._collect_results_df(self.df_phase1_stats)
        self.is_fitted = True
        return self

    def plot_phase1(self):
        """
        Plots phase 1 statistics, in this case the sample ranges and the estimated control limits.
        """
        super().plot_phase1()
        chart_type = "R" if self.variability_estimator == "range" else "s"
        unit = "range" if self.variability_estimator == "range" else "Std. dev."
        plt.title(f"Phase 1 {chart_type}-chart")
        plt.ylabel(f"Within sample variation [{unit}]")

    def plot_phase2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 2 statistics, in this case the sample ranges and the estimated control limits.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        super().plot_phase2(df_phase2=df_phase2)
        chart_type = "R" if self.variability_estimator == "range" else "s"
        unit = "range" if self.variability_estimator == "range" else "Std. dev."
        plt.title(f"Phase 2 {chart_type}-chart")
        plt.ylabel(f"Within sample variation [{unit}]")

    def plot_phase1_and_2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 1 and 2 statistics, in this case the sample averages and the estimated control limits. Phase 1
        and 2 will be displayed in differing colours to easily tell them apart visually.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        super().plot_phase1_and_2(df_phase2=df_phase2)
        chart_type = "R" if self.variability_estimator == "range" else "s"
        unit = "range" if self.variability_estimator == "range" else "Std. dev."
        plt.title(f"Phase 1 and 2 {chart_type}-chart")
        plt.ylabel(f"Within sample variation [{unit}]")


class SChart(RChart):
    """
    Univariate control chart for monitoring the in-sample variation of a quality characteristic.
    Assumes normally distributed, non-autocorrelated data. The s-chart uses an estimate
    of variability based on the in-sample standard deviations contrary to the R-chart, which uses the
    range to estimate variability. The s-chart is better suited for samples of size n > 10 or 12.
    """

    def __init__(self, n_sample_size: int = 5):
        super().__init__(n_sample_size=n_sample_size, variability_estimator="std")


#  TODO implement for sample_size > 1
#    - implement MEWMA on residual PC's control chart
class MEWMAChart(BaseControlChart, ControlChartPlotMixin):
    """
    - Note, that this is a phase 2 procedure. However, the process target/mean, mu, can reasonably
      be obtained from in-control phase 1 data.
    - Typically used with individual observations, i.e. sample size n = 1, but works for n > 1, too.
    - Small values of lambda seems a good default choice for this procedure, as it makes the procedure:
       1) robust to the underlying distribution, e.g. it is robust to data exhibiting non-normality.
       2) effective in detecting small shifts for small values of lambda.
    - It is recommended to use the EWMA together with a Shewhart chart, e.g. x-bar or Hotelling T^2,
      as the EWMA is slower at detecting large shifts than the Shewhart type chart.
    Based on Montgomery, 2013, chapter 11.4, p. 524.
    """

    def __init__(
        self,
        n_sample_size: int = 1,
        lambda_: float = 0.1,
        sigma: Optional[np.ndarray] = None,
    ):
        super().__init__(n_sample_size=n_sample_size)
        assert self.n_sample_size > 0, "Sample/subgroup size must be greater than 0."
        assert (
            0.0 < lambda_ <= 1.0
        ), "Bad lambda value given. Lambda must be in the interval 0 < lambda <= 1."
        self.stat_name = "T2"
        self.df_phase2_stats = None
        self.input_dim = None
        self.lambda_ = lambda_
        self.sigma = sigma
        self.df_PCs = None

    def fit(self, df_phase2: pd.DataFrame, verbose: bool = False, *args, **kwargs):
        self.input_dim = df_phase2.shape[1]
        assert self.input_dim > 1, (
            "Multivariate method: number of features must be more than 1. Use the EWMA procedure "
            "for univariate data."
        )
        df_phase2_copy = df_phase2.copy()

        if self.sigma is not None:
            assert self.sigma.shape == (self.input_dim, self.input_dim), (
                'The matrix "sigma" must have dimensions '
                "equal to the number of input features. "
                "E.g. 2 x 2 for 2 input features."
            )
        else:
            self.sigma = df_phase2_copy.corr()  # the correlation matrix

        Z = [np.zeros(shape=self.input_dim)]
        T2 = []
        df_phase2_copy = np.array(df_phase2_copy)
        N_rows = df_phase2_copy.shape[0]
        for i in range(1, N_rows + 1):
            if verbose and i % np.ceil(0.1 * N_rows) == 0:
                print(f"Progress: {100 * i / (df_phase2_copy.shape[0] + 1):.2f} %")
            x_i = df_phase2_copy[
                i - 1,
            ]
            z_i = self.lambda_ * x_i + (1 - self.lambda_) * Z[i - 1]  # eq: 11.30
            Z.append(z_i)
            sigma_i = (
                (self.lambda_ / (2 - self.lambda_))
                * (1 - (1 - self.lambda_) ** (2 * i))
                * self.sigma
            )  # eq: 11.32
            T2.append(
                multiply_matrices(z_i.transpose(), np.linalg.inv(sigma_i), z_i)
            )  # eq: 11.31

        Z.pop(0)
        df_Z = pd.DataFrame(
            Z,
            columns=[f"Z{i}" for i in range(1, self.input_dim + 1)],
            index=df_phase2.index,
        )
        self.df_phase2_stats = df_phase2.copy()
        self.df_phase2_stats = pd.concat([self.df_phase2_stats, df_Z], axis=1)
        self.df_phase2_stats[self.stat_name] = T2
        self.is_fitted = True
        return self

    def fit_on_PCs(
        self,
        df_phase1: pd.DataFrame,
        df_phase2: pd.DataFrame,
        n_components: int = None,
        PC_variance_explained_min: float = 0.9,
        verbose: bool = False,
        *args,
        **kwargs,
    ):
        df_transformed, pca, scaler = standardize_and_PCA(
            df_phase1, n_components=n_components
        )
        if n_components is None:
            cumulative_variances = np.cumsum(pca.explained_variance_ratio_)
            n_components = (
                np.where(cumulative_variances > PC_variance_explained_min)[0][0] + 1
            )
        assert n_components > 1, (
            "Multivariate method: number of features must be more than 1. "
            "Use the EWMA procedure for univariate data."
        )
        if verbose:
            print(f"Number of PC's used: {n_components}")
        df_phase2_transformed = apply_standardize_and_PCA(
            df_phase2, scaler=scaler, pca=pca
        )
        df_phase2_transformed = df_phase2_transformed[:, :n_components]
        # TODO try to estimate sigma on phase 1 data
        self.sigma = None  # the standardization and PCA changes the original correlations, so we can't use those
        self.df_PCs = pd.DataFrame(
            df_phase2_transformed,
            columns=[f"PC{i}" for i in range(1, n_components + 1)],
        )
        self.fit(self.df_PCs, verbose=verbose)

    def compute_delta(self, mu_shifted: np.ndarray):
        """
        This method computes the quantity, delta, of equation 11.33 of Montgomery called the non-centrality parameter.
        It is a measure of a given shift size given by a p x 1 vector, mu, of standard deviations. E.g. for a 6-dimensional
        dataset / process, a vector mu = [1, 1, 1, 1, 1, 1] corresponds to a shift of exactly one standard deviation in each
        feature dimension. Consequently, the zero-vector represents a normal condition.

        :param mu_shifted: the p x 1 shift vector of standard deviations each feature has shifted from the process target.
        :return: the non-centrality parameter, delta = (mu^T * sigma^(-1) * mu)^(1/2). This value measures the multivariate
        size of the shift, all dimensions considered.
        """
        assert (
            self.is_fitted
        ), "No correlation matrix available. Run the fit() method first on suitable phase 2 data."
        mu_shifted_copy = mu_shifted.copy()
        if isinstance(mu_shifted, list):
            mu_shifted_copy = np.array(mu_shifted)
        return np.sqrt(
            multiply_matrices(
                mu_shifted_copy.transpose(), np.linalg.inv(self.sigma), mu_shifted_copy
            )
        )

    def plot_phase2(self):
        """
        Plots the obtained phase 2 statistics for the multivariate EWMA procedure. Requires fit() to have been run first.
        """
        assert (
            self.is_fitted
        ), "No stats to plot. Run fit() first on suitable phase 2 data"
        self._plot_single_phase_univariate(self.df_phase2_stats)
        plt.title(rf"Phase 2 MEWMA-chart, $\lambda = {self.lambda_}$")
        plt.ylabel(r"Sample Hotelling $T^2_i$")
        plt.xlabel("Sample [i]")


# TODO: adapt to n_sample_size > 1 at some point
class HotellingT2Chart(BaseControlChart, ControlChartPlotMixin):
    def __init__(self, n_sample_size: int = 1, alpha: float = 0.05):
        super().__init__(n_sample_size=n_sample_size)
        self.alpha = alpha
        self.stat_name = "T2"
        self.input_dim = None
        self.sigma = None
        self.sigma_inverse = None
        self.x_bar = None
        self.m_samples = None
        self.df_phase1_stats = None
        self.df_contributions = None

    def fit(
        self,
        df_phase1: pd.DataFrame,
        compute_contributions: bool = True,
        verbose=False,
        *args,
        **kwargs,
    ):
        self.m_samples = df_phase1.shape[0]
        self.input_dim = df_phase1.shape[1]
        x = np.array(df_phase1)
        self.x_bar = np.mean(x, axis=0)
        self.sigma = np.cov(
            x.T, ddof=1
        )
        self.sigma_inverse = np.linalg.inv(self.sigma)
        T2 = []
        for i in range(x.shape[0]):
            T2.append(self._compute_T2_value(x[i]))

        self.UCL = self._compute_T2_UCL(phase=1)
        self.df_phase1_stats = pd.DataFrame(
            dict(
                T2=T2,
                UCL=self.UCL,
                outside_CL=T2 > self.UCL,
                cumulated_prop_outside_CL=np.cumsum(1 * (T2 > self.UCL))
                / self.m_samples,
            )
        )
        self.is_fitted = True

        if compute_contributions and self.input_dim > 2:
            # compute contributions by removing the k'th variable, to see how much it contributes to the T^2 value of a sample
            contributions = self._compute_T2_contributions(T2, x)
            self.df_contributions = pd.DataFrame(
                contributions, columns=df_phase1.columns, index=df_phase1.index
            )

        return self

    def _compute_T2_contributions(self, T2, x):
        contributions = np.zeros_like(x, dtype=float)
        for k in range(self.input_dim):
            x_k = np.delete(x, k, axis=1)
            x_bar_k = np.delete(self.x_bar, k)
            sigma_inverse_k = np.delete(np.delete(self.sigma, k, axis=0), k, axis=1)
            sigma_inverse_k = np.linalg.inv(sigma_inverse_k)
            for i in range(x_k.shape[0]):
                contributions[i, k] = T2[
                    i
                ] - self._compute_T2_value_given_xbar_and_sigma_inv(
                    x_k[i], x_bar_k, sigma_inverse_k
                )
        return contributions

    def predict(self, df_phase2: pd.DataFrame):
        """
        Calculates T^2 statistics for new data and returns a dataframe with the results.
        n0te: may need a better name for this method

        :param df_phase2: dataframe with new data
        :return: dataframe with the calculated T^2 results and control limits
        """
        T2 = [self._compute_T2_value(x) for _, x in df_phase2.iterrows()]
        UCL = self._compute_T2_UCL(phase=2)
        return pd.DataFrame(
            dict(
                T2=T2,
                UCL=UCL,
                outside_CL=T2 > UCL,
                cumulated_prop_outside_CL=np.cumsum(1 * (T2 > UCL)) / self.m_samples,
            )
        )

    def _compute_T2_value(self, x: np.ndarray):
        assert (
            x.shape == self.x_bar.shape
        ), f"The shape of x ({x.shape}) must be the same as x_bar: {self.x_bar.shape}"
        return multiply_matrices(
            (x - self.x_bar).T, self.sigma_inverse, (x - self.x_bar)
        )

    @staticmethod
    def _compute_T2_value_given_xbar_and_sigma_inv(
        x: np.ndarray, x_bar: np.ndarray = None, sigma_inverse: np.ndarray = None
    ):
        assert (
            x.shape == x_bar.shape
        ), f"The shape of x ({x.shape}) must be the same as x_bar: {x_bar.shape}"
        return multiply_matrices((x - x_bar).T, sigma_inverse, (x - x_bar))

    def _compute_T2_UCL(self, phase: Union[int, str]):
        p, m, _ = self.input_dim, self.m_samples, self.n_sample_size
        if int(phase) == 1:
            percentile = beta.ppf(q=1 - self.alpha, a=p / 2, b=(m - p - 1) / 2)
            self.UCL = (m - 1) ** 2 / m * percentile
        else:
            percentile = f.ppf(q=1 - self.alpha, dfn=p, dfd=m - p)
            self.UCL = (p * (m + 1) * (m - 1)) / (m * m - m * p) * percentile
        return self.UCL

    def plot_phase1(self):
        """
        Plots phase 1 statistics, in this case the sample averages and the estimated control limits.
        """
        self._plot_single_phase_univariate(self.df_phase1_stats)
        final_proportion_outside_CL = (
            self.df_phase1_stats["cumulated_prop_outside_CL"].tail(1).squeeze()
        )
        plt.suptitle(
            f"Phase 1 Hotelling $T^2$-chart,\n $\\alpha$ = {100 * self.alpha:.2f} %, "
            f"samples outside CL = {100 * final_proportion_outside_CL:.2f} %"
        )
        plt.ylabel(r"Sample $T^2$")
        plt.xlabel("Sample")

    def plot_phase2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 2 statistics, in this case the sample averages and the estimated control limits.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        df_phase2_stats = self.predict(df_phase2)
        self._plot_single_phase_univariate(df_phase2_stats)
        final_proportion_outside_CL = (
            df_phase2_stats["cumulated_prop_outside_CL"].tail(1).squeeze()
        )
        plt.suptitle(
            f"Phase 2 Hotelling $T^2$-chart,\n $\\alpha$ = {100 * self.alpha:.2f} %, "
            f"samples outside CL = {100 * final_proportion_outside_CL:.2f} %"
        )
        plt.ylabel(r"Sample $T^2$")
        plt.xlabel("Sample")

    def plot_phase1_and_2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 1 and 2 statistics, in this case the sample averages and the estimated control limits. Phase 1
        and 2 will be displayed in differing colours to easily tell them apart visually.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """

        df_phase2_results = self.predict(df_phase2)
        self._plot_two_phases(self.df_phase1_stats, df_phase2_results)
        plt.title(r"Phase 1 and 2 $\bar{X}$-chart")
        plt.ylabel(r"Sample average [$\bar{x}$]")


class PCAModelChart(HotellingT2Chart):
    def __init__(
        self,
        n_sample_size: int = 1,
        alpha: float = 0.05,
        combine_T2_and_Q: bool = False,
    ):
        super().__init__(n_sample_size=n_sample_size, alpha=alpha)
        self.scaler = None
        self.PCA = None
        self.UCL_Q = None
        self.n_components_to_retain = None
        self.loadings = None
        self.scores = None
        self.df_T2_contributions = None
        self.df_Q_contributions = None
        self.combine_T2_and_Q = combine_T2_and_Q

    def fit(
        self,
        df_phase1: pd.DataFrame,
        n_components_to_retain: int = None,
        PC_variance_explained_min: float = 0.9,
        # compute_contributions: bool = True,
        verbose=False,
        *args,
        **kwargs,
    ):
        self.n_components_to_retain = n_components_to_retain
        self.m_samples = df_phase1.shape[0]
        df_PCA_transformed, self.PCA, self.scaler = standardize_and_PCA(df_phase1)
        self.stat_name = (self.stat_name, "Q")
        if self.n_components_to_retain is None:
            (
                self.n_components_to_retain,
                cumulative_variances,
            ) = get_num_of_PCs_to_retain(self.PCA, PC_variance_explained_min)
            vprint(
                verbose,
                f"PC's used: {self.n_components_to_retain}\nData variation explained: "
                f"{100 * cumulative_variances[self.n_components_to_retain - 1]:.2f} %",
            )

        # first call parent class to estimate T^2 statistics
        df_PCA_transformed = df_PCA_transformed[:, : self.n_components_to_retain]
        PC_names = [f"PC{i}" for i in range(1, df_PCA_transformed.shape[1] + 1)]
        super().fit(
            df_phase1=pd.DataFrame(
                df_PCA_transformed, columns=PC_names, index=df_phase1.index
            )
        )
        self.is_fitted = False  # overrides is_fitted attribute set by parent's fit
        self.df_phase1_stats = self.df_phase1_stats.rename(
            columns={
                "UCL": "UCL_T2",
                "outside_CL": "outside_CL_T2",
                "cumulated_prop_outside_CL": "cumulated_prop_outside_CL_T2",
            }
        )

        # 2nd, calculate Q statistic and its control limit
        Q = self._compute_Q_values(
            df_phase1
        )  # PCA is done inside this function, so Q is done on the PC's
        self.df_phase1_stats["Q"] = Q
        self.UCL_Q = self._compute_Q_UCL(Q)
        if self.combine_T2_and_Q:
            TQ = (
                np.sqrt(np.exp(-self.UCL / self.df_phase1_stats["T2"]))
                + np.sqrt(np.exp(-self.UCL_Q / Q))
            ) / 2
            TQ_UCL = 0.6065  # from equating the statistics and their UCL's
            Q, self.UCL_Q = TQ, TQ_UCL
        df_Q_stats = pd.DataFrame(
            dict(
                UCL_Q=self.UCL_Q,
                outside_CL_Q=Q > self.UCL_Q,
                cumulated_prop_outside_CL_Q=np.cumsum(1 * (Q > self.UCL_Q))
                / df_phase1.shape[0],
            )
        )
        self.df_phase1_stats = pd.concat([self.df_phase1_stats, df_Q_stats], axis=1)
        self.df_phase1_stats.index = df_phase1.index

        ## TODO fix the contribution calculations
        ## DEPRECATED for now
        # if compute_contributions:
        #     Q_contributions, T2_contributions = self._compute_contributions(df_PCA_transformed)
        #     self.df_T2_contributions = pd.DataFrame(
        #         T2_contributions, columns=df_phase1.columns, index=df_phase1.index
        #     )
        #     self.df_Q_contributions = pd.DataFrame(
        #         Q_contributions, columns=df_phase1.columns, index=df_phase1.index
        #     )

        self.is_fitted = True

        self.hyperparameters = {
            "n_sample_size": self.n_sample_size,
            "n_components_to_retain": self.n_components_to_retain,
            "alpha": self.alpha,
            "PC_variance_explained_min": PC_variance_explained_min,
            # "compute_contributions": compute_contributions,
        }

        return self

    def predict(self, df_phase2: pd.DataFrame):
        """
        Uses the estimated phase 1 statistics from the fit() method to calculate phase 2 statistics.
        A dataframe will be returned with T^2, Q, UCL's for both, indicators whether they are outside the UCL
        and cumulated proportion of points outside.

        :param df_phase2: the phase 2 data with the same column names as provided in the phase 1 data
        :return: a dataframe with phase 2 statistics
        """
        df_transformed = apply_standardize_and_PCA(df_phase2, self.scaler, self.PCA)
        df_transformed = pd.DataFrame(df_transformed[:, : self.n_components_to_retain])
        df_phase2_stats = super().predict(df_transformed)
        Q = self._compute_Q_values(df_phase2)
        if self.combine_T2_and_Q:
            TQ = (
                np.sqrt(np.exp(-self.UCL / df_phase2_stats["T2"]))
                + np.sqrt(np.exp(-self.UCL_Q / Q))
            ) / 2
            TQ_UCL = 0.6065  # from equating the statistics and their UCL's => exp(-1) = 0.6065
            Q, self.UCL_Q = TQ, TQ_UCL
        df_phase2_stats["Q"] = Q
        df_Q_stats = pd.DataFrame(
            dict(
                UCL_Q=self.UCL_Q,
                outside_CL_Q=Q > self.UCL_Q,
                cumulated_prop_outside_CL_Q=np.cumsum(1 * (Q > self.UCL_Q))
                / df_phase2.shape[0],
            )
        )
        df_phase2_stats = df_phase2_stats.rename(
            columns={
                "UCL": "UCL_T2",
                "outside_CL": "outside_CL_T2",
                "cumulated_prop_outside_CL": "cumulated_prop_outside_CL_T2",
            }
        )
        df_phase2_stats = pd.concat([df_phase2_stats, df_Q_stats], axis=1)
        df_phase2_stats.index = df_phase2.index

        return df_phase2_stats

    # def get_contributions(self, df: pd.DataFrame):
    #     """
    #     DISABLED for now due to mismatching matrix dimensions in _compute_contributions()
    #
    #     Computes contributions for an arbitrary dataframe (phase 1 or 2). Requires fit() to be run first,
    #     as principal components and scalers need to be estimated.
    #
    #     :param df: the data observations for the given phase
    #     :return: dataframe of c ontributions
    #     """
    #     assert self.is_fitted, "Model not fitted. Run fit() method first."
    #     df_transformed = apply_standardize_and_PCA(df, self.scaler, self.PCA)
    #     df_transformed = pd.DataFrame(df_transformed[:, : self.n_components_to_retain])
    #     Q_contributions, T2_contributions = self._compute_contributions(df_transformed)
    #     T2_col_names = ["T2_contribution_" + col for col in df.columns]
    #     Q_col_names = ["Q_contribution_" + col for col in df.columns]
    #     df_T2_contributions = pd.DataFrame(
    #         T2_contributions, columns=T2_col_names, index=df.index
    #     )
    #     df_Q_contributions = pd.DataFrame(
    #         Q_contributions, columns=Q_col_names, index=df.index
    #     )
    #     return pd.concat([df_T2_contributions, df_Q_contributions], axis=1)

    # def _compute_contributions(self, df_phase1):
    #     input_dim = df_phase1.shape[1]
    #     T2_contributions = np.zeros_like(df_phase1, dtype=float)
    #     Q_contributions = np.zeros_like(df_phase1, dtype=float)
    #     A_eigenvals = np.diag(
    #         self.PCA.explained_variance_[: self.n_components_to_retain]
    #     )
    #     C_loadings = np.array(self.loadings)
    #     # X = self.scaler.transform(df_phase1)
    #     X = df_phase1
    #     X_means = np.mean(X, axis=0)
    #
    #     # n0te: there's something wrong in the matrix dimensions in the below calculations.
    #     # Todo: fix it at a later point - deprecate for now
    #     T2_matrix_scaling = multiply_matrices(
    #         C_loadings, np.linalg.inv(np.sqrt(A_eigenvals)), C_loadings.T
    #     )  # p x p
    #     Q_matrix_scaling = np.eye(input_dim) - multiply_matrices(
    #         C_loadings, C_loadings.T
    #     )  # p x p
    #     for i in range(X.shape[0]):
    #         residual = X[i] - X_means  # 1 x p
    #         T2_contributions[i] = multiply_matrices(
    #             residual, T2_matrix_scaling
    #         )  # 1 x p
    #         Q_contributions[i] = multiply_matrices(residual, Q_matrix_scaling)  # 1 x p
    #     return Q_contributions, T2_contributions

    def _compute_Q_values(self, df_raw: pd.DataFrame):
        # n0te: most of the next bit is from Max's code. Original source?
        self.scores = apply_standardize_and_PCA(df_raw, self.scaler, self.PCA)[
            :, : self.n_components_to_retain
        ]
        self.loadings = self.PCA.components_.T[
            :, : self.n_components_to_retain
        ]  # loadings from phase 1 data
        X = self.scaler.transform(df_raw)
        Xhat = multiply_matrices(self.scores, self.loadings.T)
        Q = np.sum((X - Xhat) ** 2, axis=1)  # E^T x E = SPE
        return Q

    def _compute_Q_UCL(self, Q: np.ndarray, use_box_formula: bool = False):
        """
        Computes the upper control limit (UCL) on the Q-statistics (SPE).
        Can use either the formula proposed by Box (1954) or Nomikos and MacGregor (1995).

        :param Q: array of computed Q values
        :param use_box_formula: use Box formula to compute UCL or not. Defaults to True.
        :return: the calculated UCL
        """
        if use_box_formula:
            print("Box formula for Q chart UCL not implemented yet")
            pass
        else:  # use Nomikos and MacGregor UCL:
            g = np.var(Q, ddof=1) / (2 * np.mean(Q))  # scale factor
            h = (2 * np.mean(Q) ** 2) / np.var(Q, ddof=1)  # degrees of freedom
            percentile = chi2.ppf(q=1 - self.alpha, df=h)
            self.UCL_Q = g * percentile
        return self.UCL_Q

    def plot_phase1(self):
        """
        Plots phase 1 statistics, in this case the sample averages and the estimated control limits.
        """
        final_proportions_outside_CL = (
            self.df_phase1_stats[
                ["cumulated_prop_outside_CL_T2", "cumulated_prop_outside_CL_Q"]
            ]
            .tail(1)
            .squeeze()
        )
        self._plot_single_phase_multivariate(
            self.df_phase1_stats,
            subplot_titles=[
                f"$T^2$-chart, samples outside CL: "
                f"{100 * final_proportions_outside_CL[0]:.2f} %",
                f"Q-chart, samples outside CL: "
                f"{100 * final_proportions_outside_CL[1]:.2f} %",
            ],
            y_labels=["Sample $T^2$", "Sample Q"],
        ),
        plt.suptitle(f"Phase 1, $\\alpha$ = {100 * self.alpha:.2f} %")
        if not isinstance(
            self.df_phase1_stats.index, pd.core.indexes.datetimes.DatetimeIndex
        ):
            plt.xlabel("Sample")

    def plot_phase2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 2 statistics, in this case the sample averages and the estimated control limits.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        df_phase2_stats = self.predict(df_phase2)

        final_proportions_outside_CL = (
            df_phase2_stats[
                ["cumulated_prop_outside_CL_T2", "cumulated_prop_outside_CL_Q"]
            ]
            .tail(1)
            .squeeze()
        )
        self._plot_single_phase_multivariate(
            df_phase2_stats,
            subplot_titles=[
                f"$T^2$-chart, samples outside CL: "
                f"{100 * final_proportions_outside_CL[0]:.2f} %",
                f"Q-chart, samples outside CL: "
                f"{100 * final_proportions_outside_CL[1]:.2f} %",
            ],
            y_labels=["Sample $T^2$", "Sample Q"],
            y_limit_offsets=(0.8, 1.2),
        )
        plt.suptitle(f"Phase 2, $\\alpha$ = {100 * self.alpha:.2f} %")
        if not isinstance(
            self.df_phase1_stats.index, pd.core.indexes.datetimes.DatetimeIndex
        ):
            plt.xlabel("Sample")

    def plot_phase1_and_2(self, df_phase2: pd.DataFrame):
        """
        Plots phase 1 and 2 statistics, in this case the sample averages and the estimated control limits. Phase 1
        and 2 will be displayed in differing colours to easily tell them apart visually.

        :param df_phase2: a dataframe with phase 2 data. Column names must match the phase 1 data given to fit()
        """
        df_phase2_results = self.predict(df_phase2)
        T2_outlier_percentage = (
            100 * df_phase2_results["cumulated_prop_outside_CL_T2"].tail(1).squeeze()
        )
        Q_outlier_percentage = (
            100 * df_phase2_results["cumulated_prop_outside_CL_Q"].tail(1).squeeze()
        )
        T_title = f"$T^2$-chart\n$\\alpha$ = {100 * self.alpha:.2f} %, phase 2 samples outside control limit: {T2_outlier_percentage:.2f} %"
        Q_title = f"Q-chart\n$\\alpha$ = {100 * self.alpha:.2f} %, phase 2 samples outside control limit: {Q_outlier_percentage:.2f} %"
        fig, axs = self._plot_two_phases_multivariate(
            self.df_phase1_stats,
            df_phase2_results,
            subplot_titles=[T_title, Q_title],
            y_labels=["Sample $T^2$", "Sample Q"],
        )
        fig.subplots_adjust(top=0.88)
        plt.suptitle("Phase 1 and 2 PCA SPC model")
        fig.tight_layout()
        return fig, axs


# TODO:
#  - adapt to sample size > 1
class EWMAChart(BaseControlChart, ControlChartPlotMixin):
    """The exponentially weighted moving average procedure. Useful for taking autocorrelation into account and better
    than Shewhart type charts at detecting smaller shifts. The EWMA chart is typically used for individual observations,
     i.e. sample sizes = 1.
    A useful combination is an EWMA chart
    of the observations and a Shewhart chart of the residuals from an appropriate time-series model.

    """

    def __init__(
        self,
        n_sample_size: int = 1,
        L_control_limit_width: float = 2.7,
        lambda_: float = 0.1,
        mu_process_target: Optional[float] = None,
        sigma: Optional[float] = None,
    ):
        super().__init__(n_sample_size=n_sample_size)
        assert self.n_sample_size > 0, "Sample/subgroup size must be greater than 0."
        assert (
            0.0 < lambda_ <= 1.0
        ), "Bad lambda value given. Lambda must be in the interval 0 < lambda <= 1."
        self.input_name = None
        self.stat_name = "Z"
        self.df_phase2_stats = None
        self.L_control_limit_width = L_control_limit_width
        self.mu_process_target = mu_process_target
        self.lambda_ = lambda_
        self.sigma = sigma

    def fit(self, df_phase2: pd.DataFrame, *args, **kwargs):
        df_phase2_copy = df_phase2.copy()
        df_phase2_copy = get_df_with_sample_id(
            df_phase2_copy, n_sample_size=self.n_sample_size
        )
        self.input_name = df_phase2_copy.columns.values[1]

        if self.lambda_ is None:
            self.lambda_ = np.mean(
                df_phase2_copy[self.input_name]
            )  # estimated process standard deviation, if not given

        if self.sigma is not None:
            assert (
                self.sigma > 0
            ), 'Process standard deviation, "sigma", must be a positive, real, number > 0.'
        else:
            self.sigma = np.std(
                df_phase2_copy[self.input_name], ddof=1
            )  # estimated process standard deviation, if not given

        if self.mu_process_target is None:
            self.mu_process_target = np.mean(
                df_phase2_copy[self.input_name]
            )  # estimated process mean, if not given

        Z = [self.mu_process_target]
        sigma_individual = []
        for i in range(1, df_phase2_copy.shape[0] + 1):
            x_i = df_phase2_copy[self.input_name].iloc[i - 1]
            z_i = self.lambda_ * x_i + (1 - self.lambda_) * Z[i - 1]  # eq: 9.22
            Z.append(z_i)
            sigma_i = np.sqrt(
                self.lambda_ / (2 - self.lambda_) * (1 - (1 - self.lambda_) ** (2 * i))
            )  # eq: 9.24 & 9.25
            sigma_individual.append(sigma_i)

        sigma_individual = np.array(sigma_individual)
        self.LCL = (
            self.mu_process_target
            - self.L_control_limit_width * self.sigma * sigma_individual
        )
        self.UCL = (
            self.mu_process_target
            + self.L_control_limit_width * self.sigma * sigma_individual
        )
        self.center_line = self.mu_process_target

        Z.pop(0)
        self.df_phase2_stats = df_phase2.copy()
        self.df_phase2_stats[self.stat_name] = Z
        self.df_phase2_stats = self._collect_results_df(self.df_phase2_stats)

        self.is_fitted = True
        return self

    def plot_phase2(self):
        """
        Plots the obtained phase 2 statistics for the EWMA procedure. Requires fit() to have been run first.
        """
        assert (
            self.is_fitted
        ), "No stats to plot. Run fit() first on suitable phase 2 data"
        self._plot_single_phase_univariate(self.df_phase2_stats)
        plt.title(
            rf"Phase 2 EWMA-chart, $\lambda = {self.lambda_}$, L = {self.L_control_limit_width}"
        )
        plt.ylabel(r"Sample EWMA value [$Z_i$]")
        plt.xlabel("Sample [i]")


def cusum():
    pass


def pareto_chart():
    """
    See chapter 5.4, p. 208 of Montgomery
    :return:
    """
    pass


def changepoint_model():
    """
    See chapter 10.8, p. 490 og Montgomery
    :return:
    """
    pass


def plot_ACF():
    pass


def plot_contributions():
    pass



from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn.decomposition
import sklearn.preprocessing
from statsmodels.tsa.stattools import acf, pacf

from base._src.definitions import (
    LCL_color,
    UCL_color,
    get_line_width,
    get_outside_CL_marker_size,
    multivariate_phase1_color,
    multivariate_phase2_color,
    outlier_marker_color,
    single_phase_line_color, univariate_phase1_color, univariate_phase2_color,
)


def vprint(verbose: Union[bool, int], str_to_print: str, **kwargs):
    if verbose:
        print(str_to_print, **kwargs)


def flatten_list(list_of_lists):
    return [item for sublist in list_of_lists for item in sublist]


def get_df_with_sample_id(df: pd.DataFrame, n_sample_size: int = 1):
    assert n_sample_size > 0, "Sample size must be a positive integer > 0"
    df_copy = df.copy()
    n_obs = len(df)
    n_subgroups = int(np.ceil(n_obs / n_sample_size))
    sample = np.repeat([i for i in range(1, n_subgroups + 1)], n_sample_size)
    df_copy["sample_id"] = list(sample)[:n_obs]
    # rearrange columns to have sample_id as the first column, if it isn't already
    cols = df_copy.columns.values.tolist()
    if not cols[0] == "sample_id":
        cols = cols[-1:] + cols[:-1]
    return df_copy[cols]


def multiply_matrices(*arrays: np.ndarray) -> np.float64:
    """
    Helper function to multiply an arbitrary number of matrices with appropriate dimensions wrt each other.

    :param arrays: the matrices, given as numpy arrays, to be multiplied together
    :return: the matrix product
    """
    product = arrays[0]
    for array in arrays[1:]:
        product = np.matmul(product, array)
    return product


def standardize_and_PCA(df: pd.DataFrame, n_components: Optional[int] = None):
    """
    Function to standardize the data to zero mean and unit variance and subsequently apply PCA.

    :param df: the dataframe of observations to apply transforms to
    :param n_components: optional number of principal components to retain, defaults to all
    :return: the standardized and PCA transformed dataframe, and the sklearn PCA and StandardScaler objects
    """
    scaler = sklearn.preprocessing.StandardScaler().fit(df)
    scaler.scale_ = np.std(
        df, ddof=1, axis=0
    ).values  # use unbiased estimator instead of biased
    df_transformed = scaler.transform(df)
    pca = sklearn.decomposition.PCA(n_components=n_components).fit(df_transformed)
    df_transformed = pca.transform(df_transformed)
    return df_transformed, pca, scaler


def apply_standardize_and_PCA(
    df: pd.DataFrame,
    scaler: sklearn.preprocessing._data.StandardScaler,
    pca: sklearn.decomposition._pca.PCA,
):
    """
    This function should be used for new data to be transformed using an already estimated standardize and
    PCA procedure.

    :param df: the new dataframe of observations to apply transforms to
    :param scaler: sklearn StandardScaler object
    :param pca: sklearn PCA object
    :return: the new standardized and PCA transformed dataframe
    """
    df_transformed = scaler.transform(df)
    df_transformed = pca.transform(df_transformed)
    return df_transformed


def get_num_of_PCs_to_retain(
    PCA_object: sklearn.decomposition.PCA, PC_variance_explained_min: float
):
    cumulative_variances = np.cumsum(PCA_object.explained_variance_ratio_)
    num_of_PCs_to_retain = (
        np.where(cumulative_variances > PC_variance_explained_min)[0][0] + 1
    )
    return num_of_PCs_to_retain, cumulative_variances


def plot_features_acf(
    df, gridsize: Tuple[int, int] = None, nlags: int = 50, corr_type="acf"
):
    if corr_type == "acf":
        df_acf = {
            f"{label}": acf(values, nlags=nlags, fft=True)
            for label, values in df.items()
        }
    else:
        df_acf = {f"{label}": pacf(values, nlags=nlags) for label, values in df.items()}
    df_acf = pd.DataFrame(df_acf)
    col_names = df_acf.columns.values
    nrows, ncols = gridsize
    fig, axs = plt.subplots(nrows, ncols, sharex="all", sharey="all")
    col_counter = 0
    # if gridsize is just 1-dimensional, i.e. 1 row or 1 column
    if len(axs.shape) == 1:
        for plot_number in range(max(nrows, ncols)):
            col_name = col_names[col_counter]
            axs[plot_number].stem(
                df_acf[col_name], linefmt="grey", markerfmt="", bottom=0, basefmt="r--"
            )
            axs[plot_number].set_title(col_name)
            col_counter += 1
    else:
        for row in range(nrows):
            for col in range(ncols):
                col_name = col_names[col_counter]
                axs[row, col].stem(
                    df_acf[col_name],
                    linefmt="grey",
                    markerfmt="",
                    bottom=0,
                    basefmt="r--",
                )
                axs[row, col].set_title(col_name)
                col_counter += 1
    fig.suptitle("Autocorrelation" if corr_type == "acf" else "Partial autocorrelation")
    return fig


class ControlChartPlotMixin:
    def __init__(self):
        self.UCL = None
        self.LCL = None
        self.center_line = None
        self.input_name = None
        self.stat_name = None

    @staticmethod
    def _plot_scalar_or_array(
        x: Union[float, np.ndarray, List[float]],
        ax,
        color="red",
    ):
        if isinstance(x, (np.ndarray, pd.Series, list)):
            ax.plot(x, color=color, linestyle="dashed")
        else:
            ax.axhline(x, color=color, linestyle="dashed", label="Control limits")

    def _plot_single_phase_univariate(self, df, y_limit_offsets=(0.95, 1.05)):
        fig, ax = plt.subplots(1, 1)
        df_outside_CL = df.loc[df["outside_CL"], self.stat_name]
        N_samples = df.shape[0]
        ax.plot(
            df[self.stat_name],
            linestyle="-",
            marker="",
            color=single_phase_line_color,
            linewidth=get_line_width(N_samples),
            zorder=1,
        )
        ax.scatter(
            df_outside_CL.index,
            df_outside_CL,
            marker="o",
            color="red",
            s=get_outside_CL_marker_size(N_samples),
            zorder=2,
            label="_nolegend_",
        )
        legend_labels = [self.stat_name]
        if self.center_line is not None:
            ax.axhline(self.center_line, color="blue", alpha=0.7)
            legend_labels.append("Center line")
        if self.UCL is not None:
            self._plot_scalar_or_array(self.UCL, ax, color=UCL_color)
            legend_labels.append("Control limit")
        if self.LCL is not None:
            self._plot_scalar_or_array(self.LCL, ax, color=LCL_color)

        plt.legend(legend_labels, ncol=len(legend_labels))
        y_limits = ax.get_ylim()
        ax.set_ylim(y_limits[0] * y_limit_offsets[0], y_limits[1] * y_limit_offsets[1])
        return fig

    def _plot_single_phase_multivariate(
        self,
        df,
        y_limit_offsets=(0.95, 1.05),
        gridsize: Tuple[int] = None,
        subplot_titles: List[str] = None,
        y_labels: List[str] = None,
    ):

        N_samples = df.shape[0]
        number_of_plots = len(self.stat_name)
        gridsize = (number_of_plots, 1) if gridsize is None else gridsize
        fig, axs = plt.subplots(*gridsize, sharex="all")
        for i in range(number_of_plots):
            stat_to_plot = self.stat_name[i]
            LCL_to_plot = (
                df[f"LCL_{stat_to_plot}"]
                if f"LCL_{stat_to_plot}" in df.columns
                else None
            )
            UCL_to_plot = df[f"UCL_{stat_to_plot}"]
            df_outside_CL = df.loc[df[f"outside_CL_{stat_to_plot}"], stat_to_plot]
            axs[i].plot(
                df[stat_to_plot],
                linestyle="-",
                marker="",
                color=single_phase_line_color,
                linewidth=get_line_width(N_samples),
                zorder=1,
            )
            axs[i].scatter(
                df_outside_CL.index.values,
                df_outside_CL,
                s=get_outside_CL_marker_size(N_samples),
                marker="o",
                color="red",
                zorder=2,
            )
            legend_labels = [stat_to_plot]
            if self.center_line is not None:
                axs[i].axhline(self.center_line, color="blue", alpha=0.7)
                legend_labels.append("Center line")
            if UCL_to_plot is not None:
                self._plot_scalar_or_array(UCL_to_plot, axs[i], color=UCL_color)
                legend_labels.append("Control limit")
            if LCL_to_plot is not None:
                self._plot_scalar_or_array(LCL_to_plot, axs[i], color=LCL_color)
            axs[i].legend(legend_labels, ncol=len(legend_labels))
            y_limits = axs[i].get_ylim()
            axs[i].set_ylim(
                y_limits[0] * y_limit_offsets[0], y_limits[1] * y_limit_offsets[1]
            )
            title = stat_to_plot if subplot_titles is None else subplot_titles[i]
            y_label = "" if y_labels is None else y_labels[i]
            axs[i].set_ylabel(y_label)
            axs[i].set_title(title)
        return fig, axs

    def _plot_two_phases(
        self,
        df_phase1_results: pd.DataFrame,
        df_phase2_results: pd.DataFrame,
        y_limit_offsets=(0.95, 1.05),
    ):
        fig, ax = plt.subplots(1, 1)
        df = pd.concat([df_phase1_results, df_phase2_results])
        df["phase"] = 1
        df["phase"].iloc[len(df_phase1_results) :] = 2
        df = df.reset_index()
        N_samples = df.shape[0]
        df_outside_CL = df.loc[df["outside_CL"], self.stat_name]

        LCL_to_plot = df["LCL"] if "LCL" in df.columns else None
        UCL_to_plot = df["UCL"]

        legend_labels = ["Phase 1", "Phase 2"]
        plt.plot(
            df[self.stat_name][df["phase"] == 1],
            linestyle="-",
            marker="",
            color=univariate_phase1_color,
            linewidth=get_line_width(N_samples),
            zorder=1,
        )
        plt.plot(
            df[self.stat_name][df["phase"] == 2],
            color=univariate_phase2_color,
            linewidth=get_line_width(N_samples),
            linestyle="-",
            marker="",
            zorder=1,
        )
        if self.center_line is not None:
            ax.axhline(self.center_line, color="blue", alpha=0.7)
            legend_labels.append("Center line")
        if UCL_to_plot is not None:
            self._plot_scalar_or_array(UCL_to_plot, ax, color=UCL_color)
            legend_labels.append("Control limit")
        if LCL_to_plot is not None:
            self._plot_scalar_or_array(LCL_to_plot, ax, color=LCL_color)

        ax.scatter(
            df_outside_CL.index.values,
            df_outside_CL,
            s=get_outside_CL_marker_size(N_samples),
            marker="o",
            color="red",
            zorder=2,
        )
        plt.legend(legend_labels, ncol=len(legend_labels))
        y_limits = ax.get_ylim()
        ax.set_ylim(y_limits[0] * y_limit_offsets[0], y_limits[1] * y_limit_offsets[1])
        return fig

    def _plot_two_phases_multivariate(
        self,
        df_phase1_results: pd.DataFrame,
        df_phase2_results: pd.DataFrame,
        y_limit_offsets=(0.95, 1.05),
        gridsize: Tuple[int] = None,
        subplot_titles: List[str] = None,
        y_labels: List[str] = None,
    ):

        index_is_datetime_format = isinstance(
            self.df_phase1_stats.index, pd.core.indexes.datetimes.DatetimeIndex
        )
        df = pd.concat([df_phase1_results, df_phase2_results])
        df["phase"] = 1
        df["phase"].iloc[len(df_phase1_results):] = 2
        if not index_is_datetime_format:
            df = df.reset_index()
        N_samples = df.shape[0]

        number_of_plots = len(self.stat_name)
        gridsize = (number_of_plots, 1) if gridsize is None else gridsize
        fig, axs = plt.subplots(*gridsize, sharex="all")
        for i in range(number_of_plots):
            stat_to_plot = self.stat_name[i]
            LCL_to_plot = (
                df[f"LCL_{stat_to_plot}"]
                if f"LCL_{stat_to_plot}" in df.columns
                else None
            )
            UCL_to_plot = df[f"UCL_{stat_to_plot}"]
            df_outside_CL = df.loc[df[f"outside_CL_{stat_to_plot}"], stat_to_plot]
            axs[i].plot(
                df[stat_to_plot][df["phase"] == 1],
                linestyle="-",
                marker="",
                color=multivariate_phase1_color,
                linewidth=get_line_width(N_samples),
                zorder=1,
                label="Phase 1",
            )
            axs[i].plot(
                df[stat_to_plot][df["phase"] == 2],
                linestyle="-",
                marker="",
                color=multivariate_phase2_color,
                linewidth=get_line_width(N_samples),
                zorder=1,
                label="Phase 2",
            )
            axs[i].scatter(
                df_outside_CL.index.values,
                df_outside_CL,
                s=get_outside_CL_marker_size(N_samples),
                marker="o",
                color=outlier_marker_color,
                zorder=2,
            )
            if self.center_line is not None:
                axs[i].axhline(self.center_line, color="blue", alpha=0.7)
            if UCL_to_plot is not None:
                self._plot_scalar_or_array(UCL_to_plot, axs[i], color=UCL_color)
            if LCL_to_plot is not None:
                self._plot_scalar_or_array(LCL_to_plot, axs[i], color=LCL_color)
            axs[i].legend(ncol=2)
            y_limits = axs[i].get_ylim()
            axs[i].set_ylim(
                y_limits[0] * y_limit_offsets[0], y_limits[1] * y_limit_offsets[1]
            )
            title = stat_to_plot if subplot_titles is None else subplot_titles[i]
            y_label = "" if y_labels is None else y_labels[i]
            axs[i].set_ylabel(y_label)
            axs[i].set_title(title)
            if index_is_datetime_format:
                fig.autofmt_xdate(bottom=0.1, rotation=30, ha="center")
        return fig, axs


def plot_df_acf(df, gridsize: Tuple[int, int] = None, nlags: int = 50, corr_type="acf"):
    if corr_type == "acf":
        df_acf = {
            f"{label}": acf(values, nlags=nlags, fft=True)
            for label, values in df.items()
        }
    else:
        df_acf = {f"{label}": pacf(values, nlags=nlags) for label, values in df.items()}
    df_acf = pd.DataFrame(df_acf)
    col_names = df_acf.columns.values
    nrows, ncols = gridsize
    fig, axs = plt.subplots(nrows, ncols, sharex="all", sharey="all")
    col_counter = 0
    for row in range(nrows):
        for col in range(ncols):
            col_name = col_names[col_counter]
            axs[row, col].stem(
                df_acf[col_name], linefmt="grey", markerfmt="", bottom=0, basefmt="r--"
            )
            axs[row, col].set_title(col_name)
            col_counter += 1
    fig.suptitle("Autocorrelation" if corr_type == "acf" else "Partial autocorrelation")
    return fig



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_TEP.ipynb.

# %% auto 0
__all__ = ['GetTEP']

# %% ../nbs/02_TEP.ipynb 3
import pandas as pd
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt

# %% ../nbs/02_TEP.ipynb 10
def GetTEP():
    filepath = 'https://github.com/waterboy96/TEPData/blob/1ac08a54cb9d420ff4bc0c3f0076ca06dc2ec7e4/TEP.csv?raw=true'
    TEP = pd.read_csv(filepath, index_col = [0])
    namepath = 'https://github.com/waterboy96/TEPData/blob/main/TEPnames.xlsx?raw=true'
    names = pd.read_excel(namepath, index_col = [0], header = None)
    TEP.columns = names.loc['Name']
    
    return TEP
    
    



import numpy as np
import torch
from torch.utils.data import DataLoader
from torch.utils.data.sampler import SubsetRandomSampler


class EarlyStopping:
    """Stops the training when the validation loss does not improve more than a given delta."""

    def __init__(self, delta=0, patience=10, path='best_model.pt', verbose=True):
        """
        :param delta: minimum desired improvement
        :param patience: number of successive iterations for which delta improvement must be observed
        :param path: where to save best model
        :param verbose: talkative
        """
        self.delta = delta
        self.patience = patience
        self.path = path
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_best_model(val_loss, model)

        elif score <= self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_best_model(val_loss, model)
            self.counter = 0

    def save_best_model(self, val_loss, model):
        if self.verbose:
            print(f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        torch.save(model, self.path)
        self.val_loss_min = val_loss


class TrainingModel:
    """Prepare data and train network."""

    def __init__(self, learning_rate=1e-3, batch_size=50, val_size=0.20, verbose=True):
        """
        :param penalty: importance given to the orthogonality regularization term
        :param learning_rate: to update network parameters when training
        :param batch_size: number of observations processed together before updating parameters
        :param val_size: percentage of data to be allocated to validation to avoid overfitting
        :param verbose: talkative
        """
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.val_size = val_size
        self.verbose = verbose
        self.train_loader = None
        self.valid_loader = None

    def create_datasets(self, train_data):
        """
        Creates DataLoader objects to train the autoencoder
        :param train_data: phase 1 data
        :return: dataloaders
        """

        # obtain training indices that will be used for validation
        num_train = len(train_data)
        indices = list(range(num_train))
        np.random.shuffle(indices)
        split = int(np.floor(self.val_size * num_train))
        train_idx, valid_idx = indices[split:], indices[:split]

        # define samplers for obtaining training and validation batches
        train_sampler = SubsetRandomSampler(train_idx)
        valid_sampler = SubsetRandomSampler(valid_idx)
        self.train_loader = torch.utils.data.DataLoader(train_data, batch_size=self.batch_size, sampler=train_sampler,
                                                        num_workers=0)
        self.valid_loader = torch.utils.data.DataLoader(train_data, batch_size=self.batch_size, sampler=valid_sampler,
                                                        num_workers=0)

        return self.train_loader, self.valid_loader

    def train_model(self, model, patience=10, n_epochs=100, path='best_model.pt'):
        """
        Training the model
        :param model: initialized model
        :param patience: number of epochs without improvement tolerated
        :param n_epochs: total number of epochs
        :param path: how to save best model
        :return: trained model in inference mode, train and validation loss
        """
        history = dict(train=[], val=[])
        early_stopping = EarlyStopping(patience=patience, verbose=self.verbose)
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)

        for epoch in range(1, n_epochs + 1):

            # Training
            train_losses = []
            model.train()  # prep model for training
            for batch, data in enumerate(self.train_loader, 1):
                optimizer.zero_grad()
                encoded, decoded = model(data)
                loss = model.loss_function(compression=encoded, reconstruction=decoded, original=data)
                loss.backward()
                optimizer.step()
                train_losses.append(loss.item())

            # Validation
            val_losses = []
            model.eval()  # prep model for evaluation
            with torch.no_grad():
                for data in self.valid_loader:
                    encoded, decoded = model(data)
                    loss = model.loss_function(compression=encoded, reconstruction=decoded, original=data)
                    val_losses.append(loss.item())

            # calculate average loss over an epoch
            train_loss = np.average(train_losses)
            val_loss = np.average(val_losses)
            history['train'].append(train_loss)
            history['val'].append(val_loss)
            epoch_len = len(str(n_epochs))

            if self.verbose:
                print(f'[{epoch:>{epoch_len}}/{n_epochs:>{epoch_len}}] ' + f'train_loss: {train_loss:.6f} ' +
                      f'valid_loss: {val_loss:.6f}')

            early_stopping(val_loss, model)

            if early_stopping.early_stop:
                if self.verbose:
                    print(f"-----------> Early stopping: no improvement after {patience} consecutive epochs")
                break

        model = torch.load(path)

        return model.eval(), history



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_Correlation_Matrices.ipynb.

# %% auto 0
__all__ = ['imposter_matrix', 'correlation_matrix', 'ACM', 'partial_correlation_matrix', 'PACM']

# %% ../nbs/01_Correlation_Matrices.ipynb 2
import pandas as pd
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from lag import *

# %% ../nbs/01_Correlation_Matrices.ipynb 6
def imposter_matrix(X: pd.DataFrame, # Matrix to be randomized
                   random_state = 42, # Random state to be used
                   )-> pd.DataFrame: # The input matrix with shuffled values for each column
    d = pd.DataFrame(0, index=np.arange(len(X)), columns=X.columns)
    
    for column in d:
        d[column] = X[column].sample(frac = 1, random_state=random_state).reset_index(drop=True)
        
    return d
    

# %% ../nbs/01_Correlation_Matrices.ipynb 8
def correlation_matrix(X: pd.DataFrame, # Matrix 1 for calculating the correlation matrix
                       y: pd.DataFrame, # Matrix 2 for calculating the correlation matrix
                      )-> np.ndarray: # Correlation matrix as a Numpy array
    n = np.shape(X)[0]
    
    X_scaled = (X-np.mean(X,axis = 0))/np.std(X, axis = 0, ddof = 1)
    y_scaled = (y-np.mean(y,axis = 0))/np.std(y, axis = 0, ddof = 1)
    
    correlation_matrix = np.dot(y_scaled.T,X_scaled) / (n-1)
    
    return correlation_matrix
    

# %% ../nbs/01_Correlation_Matrices.ipynb 10
def ACM(X: pd.DataFrame, # Dataframe of raw data for which to calculate the ACM
        lag: int, # Lag to calculate correlation against
        )-> pd.DataFrame: # Autocorrelation Matrix (ACM)
    
    n = np.shape(X)[0] # Number of observations
    p = np.shape(X)[1] # Number of variables
    
    X_val = X.values
    l = lag
    
    
    Xtrim = X_val[:n-l] # Observations from start to n-lag
    Xlag =  X_val[l:]   # Observations from lag to end
    
    # Calculate correlation matrix between Xtrim and Xlag
    ACM = correlation_matrix(Xtrim, Xlag)
    
    # New column names
    t0 = ['t' for i in range(p)]
    columns = [X.columns.to_list(),t0]
    
    # New index names
    t = ['t-'+f'{lag}' for i in range(p)]
    index = [X.columns.to_list(),t]
    
    ACM = pd.DataFrame(ACM)
    ACM.index = index
    ACM.columns = columns
    
    return ACM

# %% ../nbs/01_Correlation_Matrices.ipynb 12
def partial_correlation_matrix(X: pd.DataFrame, # Matrix for calculating partial correlation
                       )-> np.ndarray: # Partial correlation matrix as a Numpy array 
    V = X.cov()
    Vi = np.linalg.pinv(V)
    D = np.diag(np.sqrt(1/np.diag(Vi)))
    pcor = -1 * (D @ Vi @ D)
    pcor[np.diag_indices_from(pcor)] = 1
    
    return pcor

# %% ../nbs/01_Correlation_Matrices.ipynb 14
def PACM(X: pd.DataFrame, # Matrix for calculating partial autocorrelation
         lag: int, # Lag to calculate partial correlation against
        ):
    
    p = np.shape(X)[1]
    
    X_lag = lag_uniform(X, lag)
    
    pcor_mat = partial_correlation_matrix(X_lag)
    
    pcor_df = pd.DataFrame(pcor_mat)
    pcor_df.columns = X_lag.columns
    pcor_df.index = X_lag.columns
    
    
    PACM = pcor_df.iloc[lag*p:(lag*p+p),0:p]
    
    return PACM



"""
pytest fixtures go here
"""
import numpy as np
import pandas as pd
import pytest


@pytest.fixture()
def dataframe_for_PCAModel_phase1():
    np.random.seed(1234)
    df_phase1 = pd.DataFrame(
        {
            "x1": np.random.normal(loc=1, size=500),
            "x2": np.random.normal(loc=5, size=500),
            "x3": np.random.normal(loc=10, size=500),
        }
    )
    return df_phase1



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/dynamicPCA',
                'doc_host': 'https://waterboy96.github.io',
                'git_url': 'https://github.com/waterboy96/dynamicPCA',
                'lib_path': 'dPCA'},
  'syms': { 'dPCA.TEP': {'dPCA.TEP.GetTEP': ('tep.html#gettep', 'dPCA/TEP.py')},
            'dPCA.analysis': { 'dPCA.analysis.ACM_analysis': ('analysis.html#acm_analysis', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis': ('analysis.html#analysis', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis.__init__': ('analysis.html#analysis.__init__', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis.show_plots': ('analysis.html#analysis.show_plots', 'dPCA/analysis.py'),
                               'dPCA.analysis.PACM_analysis': ('analysis.html#pacm_analysis', 'dPCA/analysis.py')},
            'dPCA.corrmat': { 'dPCA.corrmat.ACM': ('correlation_matrices.html#acm', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.PACM': ('correlation_matrices.html#pacm', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.correlation_matrix': ('correlation_matrices.html#correlation_matrix', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.imposter_matrix': ('correlation_matrices.html#imposter_matrix', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.partial_correlation_matrix': ( 'correlation_matrices.html#partial_correlation_matrix',
                                                                           'dPCA/corrmat.py')},
            'dPCA.lag': { 'dPCA.lag.lag': ('lagfunctions.html#lag', 'dPCA/lag.py'),
                          'dPCA.lag.lag_uniform': ('lagfunctions.html#lag_uniform', 'dPCA/lag.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_LagFunctions.ipynb.

# %% auto 0
__all__ = ['lag', 'lag_uniform']

# %% ../nbs/00_LagFunctions.ipynb 2
import pandas as pd
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt

# %% ../nbs/00_LagFunctions.ipynb 6
def lag(X: pd.DataFrame, #A pandas dataframe with observations as rows and variables as columns
        lags: list, #A list with the same number of lists as the number of columns of X containing the desired lagged versions desired for the variable in that column
       )->pd.DataFrame: # A dataframe with the specified lagged variables of X

    "Return a dataframe with specific lags"
    
    res = pd.DataFrame()
    
    for x in X:
        Xlag = pd.DataFrame()
        Xlag_name = []
        Xlag_og_name = []
        for i in lags[X.columns.get_loc(x)]:
            X_shift = pd.DataFrame(X[x].shift(i))
            Xlag = pd.concat([Xlag, X_shift], axis = 1)
            
            if i == 0:
                Xlag_name.append('t')
            else:
                Xlag_name.append('t-'+f'{i}')
            Xlag_og_name.append(f'{x}')
            Xlag.columns = [Xlag_og_name,Xlag_name]
    
        res = pd.concat([res,Xlag], axis = 1)
    res = res.dropna()

    return(res)

# %% ../nbs/00_LagFunctions.ipynb 9
def lag_uniform(X: pd.DataFrame, #A pandas dataframe with observations as rows and variables as columns
        lag: int, #Integer of the maximum lag to include for all variables of X
       )->pd.DataFrame: # A dataframe with the lagged variables of X
    
    # Returns a uniformly lagged version of X

    n = np.shape(X)[0]
    p = np.shape(X)[1]
    l = lag

    res = pd.DataFrame()

    X_val = X.values

    Xlag_name = []
    Xlag_og_name = []


    for i in range(l+1):
        Xlag = pd.DataFrame(X_val[l-i:n-i])
        res = pd.concat([res,Xlag], axis = 1)
    
        if i == 0: 
            Xlag_name += ['t' for x in range(p)]
        else:
            Xlag_name += ['t-'+f'{i}' for x in range(p)]
    
        Xlag_og_name += X.columns.tolist()
    
    res.columns = [Xlag_og_name,Xlag_name]
    res.columns.names = ['Variable','Time']
    
    return res



import torch
from torch import nn
from torch.nn import functional as F


class VAE(nn.Module):

    def __init__(self, encoding_layers, penalty=0, norm_type="l2"):
        super().__init__()
        self.penalty = penalty
        self.norm_type = norm_type
        latent_dim = encoding_layers[-1]

        # Encoding layers
        encoding = []
        for i in range(len(encoding_layers) - 1):
            encoding.append(nn.Linear(encoding_layers[i], encoding_layers[i + 1]))
            encoding.append(nn.ReLU(inplace=True))
            encoding.append(nn.BatchNorm1d(encoding_layers[i + 1]))

        self.encoder = nn.Sequential(*encoding)
        self.fc_mu = nn.Linear(encoding_layers[-1], latent_dim)
        self.fc_logvar = nn.Linear(encoding_layers[-1], latent_dim)

        # Decoding layers
        decoding_layers = encoding_layers[::-1]
        decoding = []
        for i in range(len(decoding_layers) - 1):
            decoding.append(nn.Linear(decoding_layers[i], decoding_layers[i + 1]))
            decoding.append(nn.ReLU(inplace=True))
            decoding.append(nn.BatchNorm1d(decoding_layers[i + 1]))

        self.decoder = nn.Sequential(*decoding)
        self.fc_output = nn.Linear(decoding_layers[-1], encoding_layers[0])

    def encode(self, x):
        h = self.encoder(x)
        mu = self.fc_mu(h)
        log_var = self.fc_logvar(h)
        return mu, log_var

    def reparameterize(self, mu, log_var):
        std = torch.exp(0.5 * log_var)
        eps = torch.randn_like(std)
        z = mu + std * eps
        return z

    def decode(self, z):
        h = self.decoder(z)
        x_reconstructed = self.fc_output(h)
        return x_reconstructed

    def forward(self, x):
        mu, log_var = self.encode(x)
        x_encoded = self.reparameterize(mu, log_var)
        x_reconstructed = self.decode(x_encoded)
        return x_encoded, x_reconstructed

    def loss_function(self, x_encoded, x_reconstructed, original, mu, log_var):
        reconstruction_loss = F.mse_loss(x_reconstructed, original, reduction='sum')
        kl_divergence = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())
        # Compute penalty term
        dot_product = torch.matmul(x_encoded.T, x_encoded)
        if self.norm_type == 'frobenius':
            penalty_loss = torch.norm(dot_product - torch.eye(dot_product.shape[0]), p='fro') ** 2
        elif self.norm_type == 'l2':
            penalty_loss = torch.norm(dot_product - torch.eye(dot_product.shape[0]), p=2) ** 2
        else:
            raise ValueError("Invalid norm type: {}, use either frobenius or l2 norms".format(self.norm_type))
        return reconstruction_loss + kl_divergence + self.penalty * penalty_loss



from setuptools import find_packages, setup




setup(
    name="mvSPC",
    version="1.0.12",
    author="Henrik Hviid Hansen, Sebastian Olivier Nymann Topalian, Davide Cacciarelli",
    author_email="hehha@orsted.com, sebtop@kt.dtu.dk, dcac@dtu.dk",
    description="Plug and play statistical process control functions for Python",
    long_description="Python package based on work from 3 PhD projects at the Technical University of Denmark covering multivariate statistical process control (MSPC), autoencoders for MSPC and determination of delays for dynamic principal component analysis",
    long_description_content_type="markdown",
    url="https://github.com/hviidhenrik/pySPC",
    packages=find_packages(exclude=["tests*"]),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas",
        "scikit_learn",
        "scipy",
        "statsmodels",
        "torch",
        "openpyxl",
        "seaborn"
    ],
)



import pandas as pd
from autoencoders.vae.control_charts import ProcessMonitoring

training_data = pd.read_csv("datasets/tep_training.csv")
testing_data = pd.read_csv("datasets/tep_testing.csv")

SPC = ProcessMonitoring(alpha=0.05)
SPC.fit(data=training_data, encoding_layers=[33, 10], orthogonality_regularization=0, norm="frobenius", nr_epochs=1000, patience=10, batch_size=10, verbose=True)
SPC.plot_correlation_matrix()
y_train_pred = SPC.plot_phase1(log_scale=True)
y_test_pred = SPC.plot_phase2(data=testing_data, log_scale=True)
out_of_control_obs = testing_data.iloc[161, :]
SPC.spe_contribution_plots(out_of_control_obs)



"""
Unit tests go here
"""
import numpy as np
import pandas as pd
import pytest
from pandas._testing import assert_frame_equal, assert_series_equal

from base import EWMAChart, MEWMAChart, PCAModelChart, RChart, SChart, XBarChart


def sd(values):
    return np.std(values, ddof=1)


def test_xbarchart_fit_correct_values_sample_size_2_and_range():
    df_phase1 = pd.DataFrame({"x1": [1, 2, 4, 5, 8, 9]})
    fitted = XBarChart(n_sample_size=2).fit(df_phase1=df_phase1)

    expected_center_line = np.mean([1.5, 4.5, 8.5])
    constant_A2 = 1.88
    expected_Rbar = np.mean([1, 1, 1])
    expected_LCL = expected_center_line - constant_A2 * expected_Rbar
    expected_UCL = expected_center_line + constant_A2 * expected_Rbar

    assert expected_center_line == fitted.center_line
    assert expected_Rbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


def test_xbarchart_fit_correct_values_sample_size_4_and_range():
    df_phase1 = pd.DataFrame({"x1": [1, 2, 4, 1, 1, 3, 2]})
    fitted = XBarChart(n_sample_size=4, variability_estimator="range").fit(
        df_phase1=df_phase1
    )

    expected_center_line = np.mean([2, 2])
    constant_A2 = 0.729
    expected_Rbar = np.mean([3, 2])
    expected_LCL = expected_center_line - constant_A2 * expected_Rbar
    expected_UCL = expected_center_line + constant_A2 * expected_Rbar

    assert expected_center_line == fitted.center_line
    assert expected_Rbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


def test_get_df_with_sample_means_and_variability_sample_size_2_and_range():
    df_test_input = pd.DataFrame(
        {"sample_id": [1, 1, 2, 2, 3, 3, 4], "x1": [1, 2, 2, 3, 3, 4, 5]}
    )
    df_expected = pd.DataFrame(
        {"sample_mean": [1.5, 2.5, 3.5, 5], "sample_variability": [1, 1, 1, 0]},
        index=[1, 2, 3, 4],
    )
    chart = XBarChart(n_sample_size=2, variability_estimator="range").fit(df_test_input)
    df_expected.index.name = "sample_id"
    df_output = chart._get_df_with_sample_means_and_variability(
        df_test_input, grouping_column="sample_id"
    )
    assert_frame_equal(df_output, df_expected, check_dtype=False)


def test_get_df_with_sample_means_and_variability_sample_size_12_and_std():
    values = np.linspace(1, 24, 24)
    df_test_input = pd.DataFrame({"sample_id": [1] * 12 + [2] * 12, "x1": values})
    df_expected = pd.DataFrame(
        {
            "sample_mean": [6.5, 18.5],
            "sample_variability": [
                np.std(values[0:12], ddof=1),
                np.std(values[12:24], ddof=1),
            ],
        },
        index=[1, 2],
    )
    chart = XBarChart(n_sample_size=12, variability_estimator="std").fit(df_test_input)
    df_expected.index.name = "sample_id"
    df_output = chart._get_df_with_sample_means_and_variability(
        df_test_input, grouping_column="sample_id"
    )
    assert_frame_equal(df_output, df_expected, check_dtype=False)


def test_group_samples_and_compute_stats():
    df_test_input = pd.DataFrame({"x1": [1, 2, 4, 3, 2, 3, 2]})
    fitted = XBarChart(n_sample_size=2).fit(df_phase1=df_test_input)

    df_output = fitted._group_samples_and_compute_stats(df_test_input)
    df_expected = pd.DataFrame(
        {"sample_mean": [1.5, 3.5, 2.5, 2], "sample_variability": [1, 1, 1, 0]},
        index=[1, 2, 3, 4],
    )
    df_expected.index.name = "sample_id"
    assert_frame_equal(df_output, df_expected, check_dtype=False, check_index_type=False)


def test_collect_results_df_correct_output_dataframe():
    df_test_input = pd.DataFrame({"x1": [1, 2, 4, 3, 2, 3, 2]})
    fitted = XBarChart(n_sample_size=2, variability_estimator="range").fit(
        df_phase1=df_test_input
    )

    df_output = fitted._collect_results_df(fitted.df_phase1_stats)
    df_expected = pd.DataFrame(
        {
            fitted.stat_name: [1.5, 3.5, 2.5],
            "LCL": [0.62] * 3,
            "UCL": [4.38] * 3,
            "outside_CL": [False] * 3,
        },
        index=[1, 2, 3],
    )
    df_expected.index.name = "sample_id"
    assert_frame_equal(df_output, df_expected, check_dtype=False, check_index_type=False)


def test_collect_results_df_correct_bool_for_outside_CL():
    df_test_input = pd.DataFrame({"x1": [1, 2, 4, 3, 2, 3, 2, 20]})
    fitted = XBarChart(n_sample_size=2, variability_estimator="range").fit(
        df_phase1=df_test_input
    )
    fitted.LCL = 1.5
    fitted.UCL = 3.5
    df_output = fitted._collect_results_df(fitted.df_phase1_stats)

    df_expected = pd.DataFrame(
        {
            fitted.stat_name: [1.5, 3.5, 2.5, 11.0],
            "LCL": [1.5] * 4,
            "UCL": [3.5] * 4,
            "outside_CL": [True, True, False, True],
        },
        index=[1, 2, 3, 4],
    )
    df_expected.index.name = "sample_id"
    assert_frame_equal(df_output, df_expected, check_dtype=False, check_index_type=False)


def test_RChart_fit_correct_values_sample_size_2_and_range():
    df_phase1 = pd.DataFrame({"x1": [1, 3, 6, 7, 1, 3, 2]})
    fitted = RChart(n_sample_size=2, variability_estimator="range").fit(
        df_phase1=df_phase1
    )

    constant_D3 = 0
    constant_D4 = 3.267
    expected_Rbar = np.mean([2, 1, 2])
    expected_LCL = constant_D3 * expected_Rbar
    expected_UCL = constant_D4 * expected_Rbar

    assert expected_Rbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


def test_RChart_fit_correct_values_sample_size_4_and_std():
    df_phase1 = pd.DataFrame({"x1": [1, 3, 6, 7, 1, 3, 2, 4, 5]})
    fitted = RChart(n_sample_size=4, variability_estimator="std").fit(
        df_phase1=df_phase1
    )

    constant_B3 = 0
    constant_B4 = 2.266
    expected_sbar = np.mean([sd([1, 3, 6, 7]), sd([1, 3, 2, 4])])
    expected_LCL = constant_B3 * expected_sbar
    expected_UCL = constant_B4 * expected_sbar

    assert expected_sbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


def test_RChart_fit_correct_values_sample_size_4_and_range():
    df_phase1 = pd.DataFrame({"x1": [1, 2, 4, 1, 2, 5, 12]})
    fitted = RChart(n_sample_size=4, variability_estimator="range").fit(
        df_phase1=df_phase1
    )

    constant_D3 = 0
    constant_D4 = 2.282
    expected_Rbar = np.mean([3, 10])
    expected_LCL = constant_D3 * expected_Rbar
    expected_UCL = constant_D4 * expected_Rbar

    assert expected_Rbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


@pytest.mark.parametrize(
    "test_variability_estimator, test_sample_size, expected_output",
    [
        ("range", 2, "range"),
        ("Range", 30, "range"),
        ("std", 2, "std"),
        ("Std", 30, "std"),
        ("std", 2, "std"),
        ("Auto", 10, "range"),
        ("auto", 11, "std"),
    ],
)
def test_determine_varability_estimator_correct_output(
    test_variability_estimator, test_sample_size, expected_output
):
    xbarchart = XBarChart(
        n_sample_size=test_sample_size, variability_estimator=test_variability_estimator
    )
    assert expected_output == xbarchart._determine_variability_estimator(
        test_variability_estimator
    )


@pytest.mark.parametrize(
    "test_bad_variability_estimator_string",
    ["r", "a", "s", "mkaæomkm", "foo", "variance", None, -1, True, False],
)
def test_determine_varability_estimator_bad_input_fails(
    test_bad_variability_estimator_string,
):
    with pytest.raises(Exception):
        XBarChart(variability_estimator=test_bad_variability_estimator_string)


def test_remove_samples_with_only_one_observation():
    chart = XBarChart(n_sample_size=2)
    chart.input_name = "x1"
    df_test_input = pd.DataFrame({"sample_id": [1, 1, 2, 2, 3], "x1": [1] * 5})
    df_output = chart._remove_samples_with_only_one_observation(df_test_input)
    df_expected = pd.DataFrame({"sample_id": [1, 1, 2, 2], "x1": [1] * 4})
    assert_frame_equal(df_expected, df_output, check_dtype=False)


def test_schart_correct_init():
    chart = SChart(n_sample_size=2)
    assert chart.variability_estimator == "std"


def test_schart_fit_correct_values():
    df_phase1 = pd.DataFrame({"x1": [1, 3, 6, 7, 1, 3, 2, 4, 5]})
    fitted = SChart(n_sample_size=4).fit(df_phase1=df_phase1)

    constant_B3 = 0
    constant_B4 = 2.266
    expected_sbar = np.mean([sd([1, 3, 6, 7]), sd([1, 3, 2, 4])])
    expected_LCL = constant_B3 * expected_sbar
    expected_UCL = constant_B4 * expected_sbar

    assert expected_sbar == fitted.variability_Rbar_or_sbar
    assert expected_LCL == fitted.LCL
    assert expected_UCL == fitted.UCL


def test_MEWMA_fit_correct_values():
    """
    This test follows the results obtained and data used in the numerical example provided in the original
    article from 1992 by Lowry et al.: "A Multivariate Exponentially Weighted Moving Average Control Chart".
    """
    # fmt: off
    df_phase2 = pd.DataFrame(
        {
            "x1": [-1.19, 0.12, -1.69, 0.3, 0.89, 0.82, -0.3, 0.63, 1.56, 1.46],
            "x2": [0.59, 0.9, 0.4, 0.46, -0.75, 0.98, 2.28, 1.75, 1.58, 3.05],
        }
    )
    # fmt: on
    chart = MEWMAChart(lambda_=0.1, sigma=np.array([[1, 0.5], [0.5, 1]])).fit(df_phase2)
    T2_output = np.array(chart.df_phase2_stats["T2"]).round(2)
    Z1_output = np.array(chart.df_phase2_stats["Z1"]).round(2)
    Z2_output = np.array(chart.df_phase2_stats["Z2"]).round(2)

    # fmt: off
    Z1_expected = np.array([-0.12, -0.1, -0.25, -0.2, -0.09, 0.0, -0.03, 0.04, 0.19, 0.32])
    Z2_expected = np.array([0.06, 0.14, 0.17, 0.2, 0.1, 0.19, 0.4, 0.53, 0.64, 0.88])
    T2_expected = np.array([3.29, 3.18, 7.37, 5.26, 1.09, 1.28, 5.66, 8.32, 9.64, 17.21])
    # fmt: on

    assert (T2_output == T2_expected).all()
    assert (Z1_output == Z1_expected).all()
    assert (Z2_output == Z2_expected).all()


def test_MEWMA_compute_delta():
    """Using the example correlation matrix in chapter 11.4, p. 527 and the shifted mu vector of [1, 1, 1, 1, 1, 1].
    The value of delta should then be 1.86.
    """
    chart = MEWMAChart()
    chart.is_fitted = True
    # fmt: off
    chart.sigma = np.array(
        [
            [1, 0.7, 0.9, 0.3, 0.2, 0.3],
            [0.7, 1, 0.8, 0.1, 0.4, 0.2],
            [0.9, 0.8, 1, 0.1, 0.2, 0.1],
            [0.3, 0.1, 0.1, 1, 0.2, 0.1],
            [0.2, 0.4, 0.2, 0.2, 1, 0.1],
            [0.3, 0.2, 0.1, 0.1, 0.1, 1],
        ]
    )
    # fmt: on
    delta_output = np.round(chart.compute_delta(np.array([1] * 6)), 2)
    delta_expected = 1.86
    assert delta_output == delta_expected


def test_EWMA_fit_correct_values():
    """
    This test is based on the data and results of example 9.2 of Montgomery, p. 435-436.
    """
    # fmt: off
    df_phase2_input = pd.DataFrame(
        {
            "x1": [9.45, 7.99, 9.29, 11.66, 12.16, 10.18, 8.04, 11.46, 9.2, 10.34, 9.03, 11.47, 10.51, 9.4, 10.08, 9.37,
                   10.62,
                   10.31, 8.52, 10.84, 10.9, 9.33, 12.29, 11.5, 10.6, 11.08, 10.38, 11.62, 11.31, 10.52,
                   ]
        }
    )
    Z_expected = [9.94, 9.75, 9.7, 9.9, 10.13, 10.13, 9.92, 10.08, 9.99, 10.02, 9.92, 10.08, 10.12, 10.05, 10.05, 9.98,
                  10.05, 10.07,
                  9.92, 10.01, 10.1, 10.02, 10.25, 10.37, 10.4, 10.47, 10.46, 10.57, 10.65, 10.63,
                  ]
    LCL_expected = [9.73, 9.64, 9.58, 9.53, 9.5, 9.48, 9.46, 9.44, 9.43, 9.42, 9.41, 9.41, 9.4, 9.4, 9.39, 9.39, 9.39,
                    9.39, 9.39, 9.39,
                    9.38, 9.38, 9.38, 9.38, 9.38, 9.38, 9.38, 9.38, 9.38, 9.38,
                    ]
    UCL_expected = [10.27, 10.36, 10.42, 10.47, 10.5, 10.52, 10.54, 10.56, 10.57, 10.58, 10.59, 10.59, 10.6, 10.6,
                    10.61, 10.61,
                    10.61, 10.61, 10.61, 10.61, 10.62, 10.62, 10.62, 10.62, 10.62, 10.62, 10.62, 10.62, 10.62, 10.62,
                    ]
    outside_CL_expected = [False, False, False, False, False, False, False, False, False, False, False, False, False,
                           False,
                           False, False, False, False, False, False,
                           False, False, False, False, False, False, False, False, True, True,
                           ]
    # fmt: on
    chart = EWMAChart(lambda_=0.1, mu_process_target=10, sigma=1).fit(
        df_phase2=df_phase2_input
    )
    Z_output = np.array(chart.df_phase2_stats["Z"]).round(2).tolist()
    LCL_output = np.array(chart.df_phase2_stats["LCL"]).round(2).tolist()
    UCL_output = np.array(chart.df_phase2_stats["UCL"]).round(2).tolist()
    outside_CL_output = np.array(chart.df_phase2_stats["outside_CL"]).tolist()
    assert Z_output == Z_expected
    assert LCL_output == LCL_expected
    assert UCL_output == UCL_expected
    assert outside_CL_output == outside_CL_expected


def test_PCAModelChart_fit_correct_values():
    """
    This test also implicitly tests HotellingT2Chart, as the PCAModelChart uses the fit() method of the former
    to do the T^2 calculations on the principal components.
    :return:
    """
    # fmt: off
    df_phase1 = pd.DataFrame(
        {
            "x1": [1, 2, 3, 1, 2, 3, 5, 4, 3, 5, 6, 7, 5, 3, 1, 2, 3, 2, 1, 3, 2],
            "x2": [3, 4, 3, 7, 2, 8, 5, 7, 3, 5, 6, 6, 5, 6, 1, 3, 2, 1, 4, 7, 8],
            "x3": [13, 24, 33, 37, 22, 18, 25, 37, 23, 35, 36, 16, 25, 26, 11,
                   33, 22, 11, 24, 27, 28],
        }
    )
    # fmt: on
    chart = PCAModelChart(n_sample_size=1, alpha=0.05, combine_T2_and_Q=False)
    chart_combined = PCAModelChart(n_sample_size=1, alpha=0.05, combine_T2_and_Q=True)
    chart.fit(df_phase1=df_phase1, n_components_to_retain=2)
    chart_combined.fit(df_phase1=df_phase1, n_components_to_retain=2)
    output_T2 = chart.df_phase1_stats["T2"].round(3)
    output_Q = chart.df_phase1_stats["Q"].round(3)
    output_T2_UCL = chart.df_phase1_stats["UCL_T2"].mean().round(3)
    output_Q_UCL = chart.df_phase1_stats["UCL_Q"].mean().round(3)

    expected_T2_UCL = 5.393
    expected_Q_UCL = 1.852
    # fmt: off
    expected_T2 = pd.Series(
        [2.148, 0.375, 0.239, 5.399, 0.952, 0.172, 1.351, 2.283, 0.285, 1.244, 2.946, 7.842,
         1.351, 0.173, 4.095, 1.081, 0.723, 3.781, 1.400, 0.537, 1.625,
         ]
    )
    expected_Q = pd.Series(
        [0.420, 0.001, 1.260, 0.056, 0.256, 2.908, 0.016, 0.037, 0.144, 0.757, 0.554, 0.359,
         0.016, 0.181, 0.012, 0.961, 0.420, 0.001, 0.030, 0.482, 1.220,
         ]
    )
    # fmt: on
    expected_T2.name = "T2"
    expected_Q.name = "Q"
    assert_series_equal(output_T2, expected_T2)
    assert_series_equal(output_Q, expected_Q)
    assert output_T2_UCL == expected_T2_UCL
    assert output_Q_UCL == expected_Q_UCL


def test_PCAModelChart_predict_correct_values():
    # fmt: off
    df_phase1 = pd.DataFrame(
        {
            "x1": [1, 2, 3, 1, 2, 3, 5, 4, 3, 5, 6, 7, 5, 3, 1, 2, 3, 2, 1, 3, 2],
            "x2": [3, 4, 3, 7, 2, 8, 5, 7, 3, 5, 6, 6, 5, 6, 1, 3, 2, 1, 4, 7, 8],
            "x3": [13, 24, 33, 37, 22, 18, 25, 37, 23, 35, 36, 16, 25, 26, 11,
                   33, 22, 11, 24, 27, 28],
        }
    )
    # fmt: on
    chart = PCAModelChart(n_sample_size=1, alpha=0.05, combine_T2_and_Q=False)
    chart_combined = PCAModelChart(n_sample_size=1, alpha=0.05, combine_T2_and_Q=True)
    chart.fit(df_phase1=df_phase1, n_components_to_retain=2)
    chart_combined.fit(df_phase1=df_phase1, n_components_to_retain=2)
    preds = chart.predict(df_phase2=df_phase1)
    preds_combined = chart_combined.predict(df_phase2=df_phase1)

    output_T2 = preds["T2"]
    output_T2_UCL = preds["UCL_T2"][0]
    output_Q = preds["Q"]
    output_Q_UCL = preds["UCL_Q"][0]
    output_TQ = preds_combined["Q"]
    output_TQ_UCL = preds_combined["UCL_Q"][0]

    expected_T2_UCL = 7.768
    expected_Q_UCL = 1.852
    expected_TQ_UCL = 0.6065
    # fmt: off
    expected_T2 = pd.Series(
        [2.148, 0.375, 0.239, 5.399, 0.952, 0.172, 1.351, 2.283, 0.285, 1.244, 2.946, 7.842,
         1.351, 0.173, 4.095, 1.081, 0.723, 3.781, 1.400, 0.537, 1.625,
         ]
    )
    expected_Q = pd.Series(
        [0.420, 0.001, 1.260, 0.056, 0.256, 2.908, 0.016, 0.037, 0.144, 0.757, 0.554, 0.359,
         0.016, 0.181, 0.012, 0.961, 0.420, 0.001, 0.030, 0.482, 1.220,
         ]
    )
    expected_TQ = pd.Series(
        [0.325, 0.0, 0.393, 0.246, 0.161, 0.45, 0.028, 0.091, 0.061, 0.357, 0.423, 0.52, 0.028,
         0.094, 0.194, 0.378, 0.245, 0.179, 0.031, 0.267, 0.436]
    )
    # fmt: on
    expected_T2.name = "T2"
    expected_Q.name = "Q"
    expected_TQ.name = "Q"
    assert_series_equal(output_T2.round(3), expected_T2)
    assert_series_equal(output_Q.round(3), expected_Q)
    assert_series_equal(output_TQ.round(3), expected_TQ)
    assert output_T2_UCL.round(3) == expected_T2_UCL
    assert output_Q_UCL.round(3) == expected_Q_UCL
    assert output_TQ_UCL == expected_TQ_UCL


def test_compute_T2_contributions():
    # fmt: off
    df_phase1 = pd.DataFrame(
        {
            "x1": [1, 2, 3, 1, 2, 3, 5, 4, 3, 5, 6, 7, 5, 3, 1, 2, 3, 2, 1, 3, 2],
            "x2": [3, 4, 3, 7, 2, 8, 5, 7, 3, 5, 6, 6, 5, 6, 1, 3, 2, 1, 4, 7, 8],
            "x3": [13, 24, 33, 37, 22, 18, 25, 37, 23, 35, 36, 16, 25, 26, 11,
                   33, 22, 11, 24, 27, 28],
        }
    )
    # fmt: on
    df_expected = pd.DataFrame(
        dict(
            PC1=[2.139, 0.174, 0.006], PC2=[0.01, 0.2, 0.233], PC3=[0.832, 0.002, 2.497]
        )
    )
    chart = PCAModelChart(n_sample_size=1).fit(
        df_phase1=df_phase1, n_components_to_retain=3, verbose=True
    )
    df_output = chart.df_contributions.iloc[0:3].round(3)
    assert_frame_equal(df_output, df_expected, check_dtype=False)



import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler
from variational_autoencoder.training import TrainingModel
from variational_autoencoder.vae import VAE


file_path = '/Users/dcac/Data/TEP_clean_33vars/'
training_data = pd.read_csv(file_path+"train_fault_free_33columns.csv")
training_data = training_data[training_data["simulationRun"] == 1].iloc[:, 3:]
scaler = StandardScaler()
data = pd.DataFrame(scaler.fit_transform(training_data))
x = torch.Tensor(data.values)

model = VAE(encoding_layers=[33, 33])
trainer = TrainingModel(learning_rate=1e-3, batch_size=10, val_size=0.20, verbose=True)
trainer.create_datasets(train_data=x)
trained_model, losses = trainer.train_model(model, patience=10, n_epochs=1000)








